import re
from collections import ChainMap
from dataclasses import dataclass, field

# google python formatting project to save myself the trouble of being overly
# clever run the function FormatCode to transform one string into a formatted
# string
from yapf.yapflib.yapf_api import FormatCode

from .LoopIR import T
from .LoopIR import UAST, LoopIR
from .internal_cursors import Node, Gap, Block, Cursor, InvalidCursorError, GapType
from .prelude import *

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
#   Notes on Layout Schemes...

"""
  functions should return a list of strings, one for each line

  standard inputs
  tab     - holds a string of white-space
  prec    - the operator precedence of the surrounding text
            if this string contains a lower precedence operation then
            we must wrap it in parentheses.
"""

# We expect pprint to install functions on the IR rather than
# expose functions; therefore hide all variables as local
__all__ = []

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Operator Precedence

op_prec = {
    "or": 10,
    #
    "and": 20,
    #
    "<": 30,
    ">": 30,
    "<=": 30,
    ">=": 30,
    "==": 30,
    #
    "+": 40,
    "-": 40,
    #
    "*": 50,
    "/": 50,
    "%": 50,
    #
    # unary minus
    "~": 60,
}


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# UAST Pretty Printing


@extclass(UAST.proc)
@extclass(UAST.fnarg)
@extclass(UAST.stmt)
@extclass(UAST.expr)
@extclass(UAST.type)
@extclass(UAST.w_access)
def __str__(self):
    return UAST_PPrinter(self).str()


del __str__


class UAST_PPrinter:
    def __init__(self, node):
        self._node = node

        self.env = ChainMap()
        self._tab = ""
        self._lines = []

        if isinstance(node, UAST.proc):
            self.pproc(node)
        elif isinstance(node, UAST.fnarg):
            self.addline(self.pfnarg(node))
        elif isinstance(node, UAST.stmt):
            self.pstmts([node])
        elif isinstance(node, UAST.expr):
            self.addline(self.pexpr(node))
        elif isinstance(node, UAST.type):
            self.addline(self.ptype(node))
        elif isinstance(node, UAST.w_access):
            if isinstance(node, UAST.Interval):
                lo = self.pexpr(node.lo) if node.lo else "None"
                hi = self.pexpr(node.hi) if node.hi else "None"
                self.addline(f"Interval({lo},{hi})")
            elif isinstance(node, UAST.Point):
                self.addline(f"Point({self.pexpr(node.pt)})")
            else:
                assert False, "bad case"
        else:
            assert False, f"cannot print a {type(node)}"

    def str(self):
        if isinstance(self._node, (UAST.type, UAST.w_access)):
            assert len(self._lines) == 1
            return self._lines[0]

        fmtstr, linted = FormatCode("\n".join(self._lines))
        if isinstance(self._node, LoopIR.proc):
            assert linted, "generated unlinted code..."
        return fmtstr

    def push(self, only=None):
        if only is None:
            self.env = self.env.new_child()
            self._tab = self._tab + "  "
        elif only == "env":
            self.env = self.env.new_child()
        elif only == "tab":
            self._tab = self._tab + "  "
        else:
            assert False, f"BAD only parameter {only}"

    def pop(self):
        self.env = self.env.parents
        self._tab = self._tab[:-2]

    def addline(self, line):
        self._lines.append(f"{self._tab}{line}")

    def new_name(self, nm):
        strnm = str(nm)
        if strnm not in self.env:
            self.env[strnm] = strnm
            return strnm
        else:
            s = self.env[strnm]
            m = re.match("^(.*)_([0-9]*)$", s)
            # either post-pend a _1 or increment the post-pended counter
            if not m:
                s = s + "_1"
            else:
                s = f"{m[1]}_{int(m[2]) + 1}"
            self.env[strnm] = s
            return s

    def get_name(self, nm):
        strnm = str(nm)
        if strnm in self.env:
            return self.env[strnm]
        else:
            return repr(nm)

    def pproc(self, p):
        name = p.name or "_anon_"
        args = [self.pfnarg(a) for a in p.args]
        self.addline(f"def {name}({','.join(args)}):")

        self.push()
        if p.instr:
            instr_lines = p.instr.c_instr.split("\n")
            instr_lines = [f"# @instr {instr_lines[0]}"] + [
                f"#        {l}" for l in instr_lines[1:]
            ]
            for l in instr_lines:
                self.addline(l)
        for pred in p.preds:
            self.addline(f"assert {self.pexpr(pred)}")
        self.pstmts(p.body)
        self.pop()

    def pfnarg(self, a):
        mem = f" @{a.mem.name()}" if a.mem else ""
        return f"{self.new_name(a.name)} : {self.ptype(a.type)}{mem}"

    def pstmts(self, body):
        for stmt in body:
            if isinstance(stmt, UAST.Pass):
                self.addline("pass")
            elif isinstance(stmt, UAST.Assign) or isinstance(stmt, UAST.Reduce):
                op = "=" if isinstance(stmt, UAST.Assign) else "+="

                rhs = self.pexpr(stmt.rhs)

                if len(stmt.idx) > 0:
                    idx = [self.pexpr(e) for e in stmt.idx]
                    lhs = f"{self.get_name(stmt.name)}[{','.join(idx)}]"
                else:
                    lhs = self.get_name(stmt.name)

                self.addline(f"{lhs} {op} {rhs}")
            elif isinstance(stmt, LoopIR.WriteConfig):
                cname = stmt.config.name()
                rhs = self.pexpr(stmt.rhs)
                self.addline(f"{cname}.{stmt.field} = {rhs}")
            elif isinstance(stmt, UAST.FreshAssign):
                rhs = self.pexpr(stmt.rhs)
                self.addline(f"{self.new_name(stmt.name)} = {rhs}")
            elif isinstance(stmt, UAST.Alloc):
                mem = f" @{stmt.mem.name()}" if stmt.mem else ""
                self.addline(
                    f"{self.new_name(stmt.name)} : {self.ptype(stmt.type)}{mem}"
                )
            elif isinstance(stmt, UAST.Call):
                pname = stmt.f.name or "_anon_"
                args = [self.pexpr(a) for a in stmt.args]
                self.addline(f"{pname}({','.join(args)})")
            elif isinstance(stmt, UAST.If):
                cond = self.pexpr(stmt.cond)
                self.addline(f"if {cond}:")
                self.push()
                self.pstmts(stmt.body)
                self.pop()
                if len(stmt.orelse) > 0:
                    self.addline("else:")
                    self.push()
                    self.pstmts(stmt.orelse)
                    self.pop()
            elif isinstance(stmt, UAST.For):
                cond = self.pexpr(stmt.cond)
                self.push(only="env")
                self.addline(f"for {self.new_name(stmt.iter)} in {cond}:")
                self.push(only="tab")
                self.pstmts(stmt.body)
                self.pop()
            else:
                assert False, "unrecognized stmt type"

    def pexpr(self, e, prec=0):
        if isinstance(e, UAST.Read):
            if len(e.idx) > 0:
                idx = [self.pexpr(i) for i in e.idx]
                return f"{self.get_name(e.name)}[{','.join(idx)}]"
            else:
                return self.get_name(e.name)
        elif isinstance(e, UAST.Const):
            return str(e.val)
        elif isinstance(e, UAST.BinOp):
            local_prec = op_prec[e.op]
            # increment rhs by 1 to account for left-associativity
            lhs = self.pexpr(e.lhs, prec=local_prec)
            rhs = self.pexpr(e.rhs, prec=local_prec + 1)
            s = f"{lhs} {e.op} {rhs}"
            # if we have a lower precedence than the environment...
            if local_prec < prec:
                s = f"({s})"
            return s
        elif isinstance(e, UAST.USub):
            return f"-{self.pexpr(e.arg, prec=op_prec['~'])}"
        elif isinstance(e, UAST.ParRange):
            return f"par({self.pexpr(e.lo)},{self.pexpr(e.hi)})"
        elif isinstance(e, UAST.SeqRange):
            return f"seq({self.pexpr(e.lo)},{self.pexpr(e.hi)})"
        elif isinstance(e, UAST.WindowExpr):

            def pacc(w):
                if isinstance(w, UAST.Point):
                    return self.pexpr(w.pt)
                elif isinstance(w, UAST.Interval):
                    lo = self.pexpr(w.lo) if w.lo else ""
                    hi = self.pexpr(w.hi) if w.hi else ""
                    return f"{lo}:{hi}"
                else:
                    assert False, "bad case"

            return f"{self.get_name(e.name)}[{', '.join([pacc(w) for w in e.idx])}]"
        elif isinstance(e, UAST.StrideExpr):
            return f"stride({self.get_name(e.name)}, {e.dim})"
        elif isinstance(e, UAST.BuiltIn):
            pname = e.f.name() or "_anon_"
            args = [self.pexpr(a) for a in e.args]
            return f"{pname}({','.join(args)})"
        elif isinstance(e, LoopIR.ReadConfig):
            cname = e.config.name()
            return f"{cname}.{e.field}"
        else:
            assert False, "unrecognized expr type"

    def ptype(self, t):
        if isinstance(t, UAST.Num):
            return "R"
        elif isinstance(t, UAST.F16):
            return "f16"
        elif isinstance(t, UAST.F32):
            return "f32"
        elif isinstance(t, UAST.F64):
            return "f64"
        elif isinstance(t, UAST.INT8):
            return "i8"
        elif isinstance(t, UAST.UINT8):
            return "ui8"
        elif isinstance(t, UAST.UINT16):
            return "ui16"
        elif isinstance(t, UAST.INT32):
            return "i32"
        elif isinstance(t, UAST.Bool):
            return "bool"
        elif isinstance(t, UAST.Int):
            return "int"
        elif isinstance(t, UAST.Index):
            return "index"
        elif isinstance(t, UAST.Size):
            return "size"
        elif isinstance(t, UAST.Tensor):
            base = str(t.basetype())
            if t.is_window:
                base = f"[{base}]"
            rngs = ",".join([self.pexpr(r) for r in t.shape()])
            return f"{base}[{rngs}]"
        else:
            assert False, "impossible type case"


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# LoopIR Pretty Printing


def _format_code(code):
    return FormatCode(code)[0].rstrip("\n")


@extclass(LoopIR.proc)
def __str__(self):
    return _format_code("\n".join(_print_proc(self, PrintEnv(), "")))


@extclass(LoopIR.fnarg)
def __str__(self):
    return _format_code(_print_fnarg(self, PrintEnv()))


@extclass(LoopIR.stmt)
def __str__(self):
    return _format_code("\n".join(_print_stmt(self, PrintEnv(), "")))


@extclass(LoopIR.expr)
def __str__(self):
    return _format_code(_print_expr(self, PrintEnv()))


@extclass(LoopIR.type)
def __str__(self):
    return _format_code(_print_type(self, PrintEnv()))


del __str__


@dataclass
class PrintEnv:
    env: ChainMap[Sym, str] = field(default_factory=ChainMap)
    names: ChainMap[str, int] = field(default_factory=ChainMap)

    def push(self) -> "PrintEnv":
        return PrintEnv(self.env.new_child(), self.names.new_child())

    def get_name(self, nm):
        if resolved := self.env.get(nm):
            return resolved

        candidate = str(nm)
        num = self.names.get(candidate, 1)
        while candidate in self.names:
            candidate = f"{nm}_{num}"
            num += 1

        self.env[nm] = candidate
        self.names[str(nm)] = num
        return candidate


def _print_proc(p, env: PrintEnv, indent: str) -> list[str]:
    args = [_print_fnarg(a, env) for a in p.args]

    lines = [f"{indent}def {p.name}({', '.join(args)}):"]

    indent = indent + "  "

    if p.instr:
        for i, line in enumerate(p.instr.c_instr.split("\n")):
            if i == 0:
                lines.append(f"{indent}# @instr {line}")
            else:
                lines.append(f"{indent}#        {line}")

    for pred in p.preds:
        lines.append(f"{indent}assert {_print_expr(pred, env)}")

    lines.extend(_print_block(p.body, env, indent))

    return lines


def _print_block(blk, env: PrintEnv, indent: str) -> list[str]:
    lines = []
    for stmt in blk:
        lines.extend(_print_stmt(stmt, env, indent))
    return lines


def _print_stmt(stmt, env: PrintEnv, indent: str) -> list[str]:
    if isinstance(stmt, LoopIR.Pass):
        return [f"{indent}pass"]

    elif isinstance(stmt, (LoopIR.Assign, LoopIR.Reduce)):
        op = "=" if isinstance(stmt, LoopIR.Assign) else "+="

        lhs = env.get_name(stmt.name)

        idx = [_print_expr(e, env) for e in stmt.idx]
        idx = f"[{', '.join(idx)}]" if idx else ""

        rhs = _print_expr(stmt.rhs, env)

        return [f"{indent}{lhs}{idx} {op} {rhs}"]

    elif isinstance(stmt, LoopIR.WriteConfig):
        cname = stmt.config.name()
        rhs = _print_expr(stmt.rhs, env)
        return [f"{indent}{cname}.{stmt.field} = {rhs}"]

    elif isinstance(stmt, LoopIR.WindowStmt):
        rhs = _print_expr(stmt.rhs, env)
        return [f"{indent}{env.get_name(stmt.name)} = {rhs}"]

    elif isinstance(stmt, LoopIR.Alloc):
        mem = f" @{stmt.mem.name()}" if stmt.mem else ""
        ty = _print_type(stmt.type, env)
        return [f"{indent}{env.get_name(stmt.name)} : {ty}{mem}"]

    elif isinstance(stmt, LoopIR.Free):
        # mem = f"@{stmt.mem.name()}" if stmt.mem else ""
        return [f"{indent}free({env.get_name(stmt.name)})"]

    elif isinstance(stmt, LoopIR.Call):
        args = [_print_expr(a, env) for a in stmt.args]
        return [f"{indent}{stmt.f.name}({', '.join(args)})"]

    elif isinstance(stmt, LoopIR.If):
        cond = _print_expr(stmt.cond, env)
        lines = [f"{indent}if {cond}:"]
        lines.extend(_print_block(stmt.body, env.push(), indent + "  "))
        if stmt.orelse:
            lines.append(f"{indent}else:")
            lines.extend(_print_block(stmt.orelse, env.push(), indent + "  "))
        return lines

    elif isinstance(stmt, LoopIR.For):
        lo = _print_expr(stmt.lo, env)
        hi = _print_expr(stmt.hi, env)
        body_env = env.push()
        loop_type = "par" if isinstance(stmt.loop_mode, LoopIR.Par) else "seq"
        lines = [
            f"{indent}for {body_env.get_name(stmt.iter)} in {loop_type}({lo}, {hi}):"
        ]
        lines.extend(_print_block(stmt.body, body_env, indent + "  "))
        return lines

    assert False, f"unrecognized stmt: {type(stmt)}"


def _print_fnarg(a, env: PrintEnv) -> str:
    if a.type == T.size:
        return f"{env.get_name(a.name)} : size"
    elif a.type == T.index:
        return f"{env.get_name(a.name)} : index"
    else:
        ty = _print_type(a.type, env)
        mem = f" @{a.mem.name()}" if a.mem else ""
        return f"{env.get_name(a.name)} : {ty}{mem}"


def _print_expr(e, env: PrintEnv, prec: int = 0) -> str:
    if isinstance(e, LoopIR.Read):
        name = env.get_name(e.name)
        idx = f"[{', '.join(_print_expr(i, env) for i in e.idx)}]" if e.idx else ""
        return f"{name}{idx}"

    elif isinstance(e, LoopIR.Const):
        return str(e.val)

    elif isinstance(e, LoopIR.USub):
        return f'-{_print_expr(e.arg, env, prec=op_prec["~"])}'

    elif isinstance(e, LoopIR.BinOp):
        local_prec = op_prec[e.op]
        # increment rhs by 1 to account for left-associativity
        lhs = _print_expr(e.lhs, env, prec=local_prec)
        rhs = _print_expr(e.rhs, env, prec=local_prec + 1)
        s = f"{lhs} {e.op} {rhs}"
        # if we have a lower precedence than the environment...
        if local_prec < prec:
            s = f"({s})"
        return s

    elif isinstance(e, LoopIR.WindowExpr):
        name = env.get_name(e.name)
        return f"{name}[{', '.join([_print_w_access(w, env) for w in e.idx])}]"

    elif isinstance(e, LoopIR.StrideExpr):
        return f"stride({env.get_name(e.name)}, {e.dim})"

    elif isinstance(e, LoopIR.BuiltIn):
        pname = e.f.name() or "_anon_"
        args = [_print_expr(a, env) for a in e.args]
        return f"{pname}({', '.join(args)})"

    elif isinstance(e, LoopIR.ReadConfig):
        cname = e.config.name()
        return f"{cname}.{e.field}"

    assert False, f"unrecognized expr: {type(e)}"


def _print_type(t, env: PrintEnv) -> str:
    if isinstance(t, T.Num):
        return "R"
    elif isinstance(t, T.F16):
        return "f16"
    elif isinstance(t, T.F32):
        return "f32"
    elif isinstance(t, T.F64):
        return "f64"
    elif isinstance(t, T.INT8):
        return "i8"
    elif isinstance(t, T.UINT8):
        return "ui8"
    elif isinstance(t, T.UINT16):
        return "ui16"
    elif isinstance(t, T.INT32):
        return "i32"
    elif isinstance(t, T.Bool):
        return "bool"
    elif isinstance(t, T.Int):
        return "int"
    elif isinstance(t, T.Index):
        return "index"
    elif isinstance(t, T.Size):
        return "size"
    elif isinstance(t, T.Error):
        return "err"

    elif isinstance(t, T.Tensor):
        base = _print_type(t.basetype(), env)
        if t.is_window:
            base = f"[{base}]"
        ranges = ", ".join([_print_expr(r, env) for r in t.shape()])
        return f"{base}[{ranges}]"

    elif isinstance(t, T.Window):
        # Below, we print idx='[x:y]' with single quotes because yapf can't
        # parse the colon in idx=[0:n] since it thinks its assignment.
        return (
            f"Window(src_type={t.src_type},as_tensor={t.as_tensor},"
            f"src_buf={t.src_buf},"
            f"idx='[{', '.join([_print_w_access(w, env) for w in t.idx])}]')"
        )
    elif isinstance(t, T.Stride):
        return "stride"

    assert False, f"impossible type {type(t)}"


def _print_w_access(node, env: PrintEnv) -> str:
    if isinstance(node, LoopIR.Interval):
        lo = _print_expr(node.lo, env)
        hi = _print_expr(node.hi, env)
        return f"{lo}:{hi}"

    elif isinstance(node, LoopIR.Point):
        return _print_expr(node.pt, env)

    assert False, "bad case"


def _print_cursor(cur):
    if isinstance(cur, Node) and not isinstance(cur._node, (LoopIR.proc, LoopIR.stmt)):
        raise NotImplementedError(
            "Cursor printing is only implemented for procs and statements"
        )

    root_cur = cur.root()
    lines = _print_cursor_proc(root_cur, cur, PrintEnv(), "")
    code = _format_code("\n".join(lines))
    # need to use "..." for Python parsing, but unquoted ellipses are prettier
    code = code.replace('"..."', "...")
    return code


def _print_cursor_proc(
    cur: Node, target: Cursor, env: PrintEnv, indent: str
) -> list[str]:
    p = cur._node
    assert isinstance(p, LoopIR.proc)

    match_comment = "  # <-- NODE" if cur == target else ""

    args = [_print_fnarg(a, env) for a in p.args]
    lines = [f"{indent}def {p.name}({', '.join(args)}):{match_comment}"]

    indent = indent + "  "

    if cur == target:
        if p.instr:
            for i, line in enumerate(p.instr.c_instr.split("\n")):
                if i == 0:
                    lines.append(f"{indent}# @instr {line}")
                else:
                    lines.append(f"{indent}#        {line}")

        for pred in p.preds:
            lines.append(f"{indent}assert {_print_expr(pred, env)}")

    lines.extend(_print_cursor_block(cur.body(), target, env, indent))
    return lines


def _print_cursor_block(
    cur: Block, target: Cursor, env: PrintEnv, indent: str
) -> list[str]:
    def if_cursor(c, move, k):
        try:
            return k(move(c))
        except InvalidCursorError:
            return []

    def more_stmts(_):
        return [f'{indent}"..."']

    def local_stmt(c):
        return _print_cursor_stmt(c, target, env, indent)

    if isinstance(target, Gap) and target in cur:
        if target._type == GapType.Before:
            return [
                *if_cursor(target, lambda g: g.anchor().prev(2), more_stmts),
                *if_cursor(target, lambda g: g.anchor().prev(), local_stmt),
                f"{indent}[GAP - Before]",
                *if_cursor(target, lambda g: g.anchor(), local_stmt),
                *if_cursor(target, lambda g: g.anchor().next(), more_stmts),
            ]
        else:
            assert target._type == GapType.After
            return [
                *if_cursor(target, lambda g: g.anchor().prev(), more_stmts),
                *if_cursor(target, lambda g: g.anchor(), local_stmt),
                f"{indent}[GAP - After]",
                *if_cursor(target, lambda g: g.anchor().next(), local_stmt),
                *if_cursor(target, lambda g: g.anchor().next(2), more_stmts),
            ]

    elif isinstance(target, Block) and target in cur:
        block = [f"{indent}# BLOCK START"]
        for stmt in target:
            block.extend(local_stmt(stmt))
        block.append(f"{indent}# BLOCK END")
        return [
            *if_cursor(target, lambda g: g[0].prev(), more_stmts),
            *block,
            *if_cursor(target, lambda g: g[-1].next(), more_stmts),
        ]

    else:
        stmt = next(filter(lambda s: s.is_ancestor_of(target), cur), None)
        if stmt is None:
            return [f'{indent}"..."']

        return [
            *if_cursor(stmt, lambda g: g.prev().before(), more_stmts),
            *local_stmt(stmt),
            *if_cursor(stmt, lambda g: g.next().after(), more_stmts),
        ]


def _print_cursor_stmt(
    cur: Node, target: Cursor, env: PrintEnv, indent: str
) -> list[str]:
    stmt = cur._node

    if isinstance(stmt, LoopIR.If):
        cond = _print_expr(stmt.cond, env)
        lines = [f"{indent}if {cond}:"]
        lines.extend(_print_cursor_block(cur.body(), target, env.push(), indent + "  "))
        if stmt.orelse:
            lines.append(f"{indent}else:")
            lines.extend(
                _print_cursor_block(cur.orelse(), target, env.push(), indent + "  ")
            )

    elif isinstance(stmt, LoopIR.For):
        lo = _print_expr(stmt.lo, env)
        hi = _print_expr(stmt.hi, env)
        body_env = env.push()
        loop_type = "par" if isinstance(stmt.loop_mode, LoopIR.Par) else "seq"
        lines = [
            f"{indent}for {body_env.get_name(stmt.iter)} in {loop_type}({lo}, {hi}):",
            *_print_cursor_block(cur.body(), target, body_env, indent + "  "),
        ]

    else:
        lines = _print_stmt(stmt, env, indent)
    if cur == target:
        lines[0] = f"{lines[0]}  # <-- NODE"

    return lines



from exo import DRAM

import exo.API_cursors as _PC
from ..API import Procedure, SchedulingError


def get_observed_stmts(stmt_cursor):
    """
    Generator of observed statement by this statement cursor

    Args:
        stmt_cursor (StmtCursor): a statement cursor

    Raises:
        TypeError: if stmt_cursor is not of the correct type

    Yields:
        StmtCursor: cursors to all the previous statements

    def proc(...):
        s1;
        s2;
        if ... : (s3)
            s4;
            s5;
        for ... : (s6) <- stmt_cursor.parent()
            s7; <- stmt_cursor.prev()
            s8; <- stmt_cursor
            s9;

        yields: s7, s6, s3; s2; s1
    """
    if not isinstance(stmt_cursor, _PC.StmtCursor):
        raise TypeError("stmt_cursor must be an instance of StmtCursor")

    cursor = stmt_cursor
    while True:
        prev_cursor = cursor.prev()

        # This means cursor was the first statement in a body
        if type(prev_cursor) is _PC.InvalidCursor:
            prev_cursor = cursor.parent()

            # This means cursor was the first statement in the proc
            if type(prev_cursor) is _PC.InvalidCursor:
                return

        cursor = prev_cursor
        yield cursor


def check_call_mem_types(call_cursor):
    """
    check memory consistency between the called procedure
    and the arguments passed at the call site.

    Args:
        call_cursor (CallCursor): call to check memory consistency at

    Raises:
        TypeError: if the cursor provided isn't a `CallCursor`

    Returns:
        bool: whether memory consistency holds at the call site
    """
    if not isinstance(call_cursor, _PC.CallCursor):
        raise TypeError("call_cursor must be a CallCursor")

    ###################################################################
    # build an env of symbols this call statement observes.
    # e.g. {x: DRAM, y: Neon}
    ###################################################################
    env = {}
    caller = call_cursor.proc()

    # Add proc parameters to env
    for arg in caller.args():
        if arg.type().is_numeric():
            mem = arg.mem()
            env[arg.name()] = mem

    # Search through observed statement to find allocations
    for stmt_cursor in get_observed_stmts(call_cursor):
        # Found a buffer allocation, record memory type
        if type(stmt_cursor) is _PC.AllocCursor:
            mem = stmt_cursor.mem()
            env[stmt_cursor.name()] = mem

    ###################################################################
    # Check memory consistency at call site
    ###################################################################
    call_args = call_cursor.args()
    callee_parameters = call_cursor.subproc().args()
    for ca, sa in zip(call_args, callee_parameters):
        if sa.type().is_numeric():
            smem = sa.mem()
            cmem = env[ca.name()]
            # Check if the argument memory type is a subclass of the callee's parameter
            if not issubclass(cmem, smem):
                return False

    return True



from __future__ import annotations
from dataclasses import dataclass

from exo import *
from exo.syntax import *
from exo.API_cursors import *

from .scheduling import *
from .inspection import *
from .higher_order import *
from .rc_wrappers import *


@dataclass
class bind_and_set_expr_cursors:
    alloc: AllocCursor
    bound_expr: ExprCursor
    expr_reads: ExprCursor


def bind_and_set_expr(proc, exprs, precision, memory, new_name=None, rc=False):
    if new_name is None:
        new_name = next(get_unique_names(proc))

    expr = exprs if isinstance(exprs, ExprCursor) else exprs[0]
    stmt = get_enclosing_stmt(proc, expr)
    proc = bind_expr(proc, exprs, new_name)
    proc = set_precision(proc, new_name, precision)
    proc = set_memory(proc, new_name, memory)

    alloc = get_declaration(proc, stmt, new_name)
    bound_expr = alloc.next().rhs()

    if not rc:
        return proc
    # Disabled since forwarding after replace is not supported now
    # exprs = [proc.forward(e) for e in exprs]
    return proc, bind_and_set_expr_cursors(alloc, bound_expr, exprs)


def parallelize_and_lift_alloc(proc, alloc_cursor, n_lifts=1):
    """
    for i in seq(0, hi):
        B1;
        name: type[shape];
        B2;

    ----->

    name: type[hi][shape]
    for i in seq(0, hi):
        B1;
        B2;
    """
    alloc_cursor = proc.forward(alloc_cursor)
    for i in range(n_lifts):
        alloc_cursor = proc.forward(alloc_cursor)
        enclosing_scope = alloc_cursor.parent()
        if isinstance(enclosing_scope, ForCursor):
            proc = expand_dim(
                proc,
                alloc_cursor,
                expr_to_string(enclosing_scope.hi()),
                enclosing_scope.name(),
            )
        proc = lift_alloc(proc, alloc_cursor)
    return proc


def parallelize_allocs(proc, cursor):
    if not isinstance(cursor, (ForCursor, IfCursor)):
        raise SchedulingError(
            f"Got type {type(cursor)}, expected {ForCursor} or {IfCursor}"
        )

    allocs = filter(lambda s: isinstance(s, AllocCursor), nlr_stmts(proc, cursor))
    func = lambda proc, alloc: parallelize_and_lift_alloc(
        proc, alloc, get_distance(proc, alloc, cursor)
    )
    return apply(func)(proc, allocs)


def interleave_loop(proc, loop, factor=None, par_reduce=False, memory=DRAM, tail="cut"):
    """
    for i in seq(0, c):
        S1
        S2
        S3

    ----->
    s1 x c
    s2 x c
    s3 x c
    """
    if factor == 1:
        return proc

    loop = proc.forward(loop)

    def rewrite(proc, loop, factor=None, par_reduce=False, memory=DRAM, tail="cut"):
        loop = proc.forward(loop)
        if factor is not None:
            proc, (outer, loop, _) = divide_loop_(
                proc, loop, factor, tail=tail, rc=True
            )
        else:
            outer = loop.parent()
        if par_reduce:
            proc = parallelize_all_reductions(proc, outer, memory=memory, unroll=True)
            loop = proc.forward(outer).body()[0]
        allocs = filter(lambda s: isinstance(s, AllocCursor), loop.body())
        proc = apply(parallelize_and_lift_alloc)(proc, allocs)

        stmts = list(proc.forward(loop).body())
        proc = apply(fission)(proc, [s.after() for s in stmts[:-1]])
        proc = apply(unroll_loop)(proc, [proc.forward(s).parent() for s in stmts])
        return proc

    if tail in {"cut", "cut_and_guard"}:
        proc = rewrite(proc, loop, factor, par_reduce, memory, tail)
    elif tail == "recursive":
        if factor is None:
            raise SchedulingError(
                "Cannot specify recursive tail strategy and factor=None"
            )
        proc, (_, inners, _) = divide_loop_recursive(
            proc, loop, factor, tail="cut", rc=True
        )
        proc = apply(rewrite)(proc, inners, par_reduce=par_reduce, memory=memory)
    elif tail == "specialize":
        if factor is None:
            raise SchedulingError(
                "Cannot specify recursive tail strategy and factor=None"
            )
        proc = rewrite(proc, loop, factor, par_reduce, memory, tail="cut")
        tail_loop = proc.forward(loop).next()
        proc, (stmts,) = binary_specialize(
            proc, tail_loop, tail_loop.hi(), [i for i in range(factor)], rc=True
        )
        proc = apply(rewrite)(proc, stmts, par_reduce=par_reduce, memory=memory)
    else:
        raise SchedulingError(f"Unknown tail strategy: {tail}")
    return proc


@dataclass
class hoist_stmt_cursors:
    allocs: list[AllocCursor]
    stmt: StmtCursor
    loop: ForCursor

    def __iter__(self):
        yield self.allocs
        yield self.stmt
        yield self.loop


def hoist_stmt(proc, stmt, rc=False):
    """
    for i in seq(0, hi):
        B1;
        s;
        B2;

    --->

    s;
    for i in seq(0, hi):
        B1;
        B2;
    """
    stmt = proc.forward(stmt)

    # Type Checking
    if not isinstance(stmt, StmtCursor):
        raise SchedulingError("Cannot hoist cursor that are not statements")

    loop = stmt.parent()

    # Pre-condition 1: a scope exists
    if not isinstance(loop, ForCursor):
        raise SchedulingError("Statement is not within a loop")

    # Pre-condition 2: fail-fast, no dependency on a loop
    deps = list(get_symbols(proc, stmt))
    if isinstance(loop, ForCursor) and loop.name() in deps:
        raise SchedulingError(
            "Cannot hoist cursor to a statement that depends on enclosing loop"
        )

    # Alloc is a special case
    if isinstance(stmt, AllocCursor):
        proc = lift_alloc(proc, stmt)
        if not rc:
            return proc
        loop = proc.forward(loop)
        stmt = proc.forward(stmt)
        return proc, hoist_stmt_cursors([stmt], stmt, loop)

    allocs = []

    # Reorder the statement to the top of the loop
    while not isinstance(stmt.prev(), InvalidCursor):
        prev_stmt = stmt.prev()
        if isinstance(prev_stmt, AllocCursor) and prev_stmt.name() in deps:
            proc = lift_alloc(proc, prev_stmt)
            allocs.append(prev_stmt)
        else:
            proc = reorder_stmts(proc, stmt.expand(1, 0))
        stmt = proc.forward(stmt)

    # Pull the statement on its own outside the loop
    if len(loop.body()) > 1:
        proc, (_, loop) = fission_(proc, stmt.after(), rc=True)
        stmt = proc.forward(stmt)
    proc = remove_loop(proc, stmt.parent())

    if not rc:
        return proc

    allocs = [proc.forward(a) for a in allocs]
    stmt = proc.forward(stmt)
    loop = proc.forward(loop)
    return proc, hoist_stmt_cursors(allocs, stmt, loop)


@dataclass
class hoist_from_loop_cursors:
    hoisted: list
    loop: ForCursor

    def __iter__(self):
        yield self.hoisted
        yield self.loop


def hoist_from_loop(proc, loop, rc=False):
    loop = proc.forward(loop)

    if not is_loop(proc, loop):
        raise SchedulingError(f"loop must of type {ForCursor} not {type(loop)}")

    rcs = []

    def hoist_non_alloc(proc, stmt):
        stmt = proc.forward(stmt)
        if isinstance(stmt, AllocCursor):
            return proc
        proc, cursors = hoist_stmt(proc, stmt, rc=True)
        rcs.append(cursors)
        return proc

    proc = apply(attempt(hoist_non_alloc))(proc, loop.body())

    if not rc:
        return proc

    if not rcs:
        return proc, hoist_from_loop_cursors([], loop)

    loop = rcs[-1].loop
    stmt_allocs = []
    for cursors in rcs:
        stmt = proc.forward(cursors.stmt)
        allocs = tuple(proc.forward(alloc) for alloc in cursors.allocs)
        stmt_allocs.append((stmt, allocs))
    return proc, hoist_from_loop_cursors(stmt_allocs, loop)


@dataclass
class jam_stmt_cursors:
    loop: ForCursor
    stmt: StmtCursor

    def __iter__(self):
        yield self.loop
        yield self.stmt


def jam_stmt(proc, stmt, unsafe_disable_check=False, rc=False):
    stmt = proc.forward(stmt)
    loop = stmt.next()
    if not is_loop(proc, loop):
        raise SchedulingError("Next statement must be a loop.")

    proc = add_loop(
        proc,
        stmt,
        loop.name(),
        FormattedExprStr("_ - _", loop.hi(), loop.lo()),
        unsafe_disable_check=unsafe_disable_check,
    )
    stmt = proc.forward(stmt)
    stmt_loop = proc.forward(stmt).parent()
    proc = shift_loop(proc, stmt_loop, FormattedExprStr("_", loop.lo()))
    proc = simplify(proc)
    proc = fuse(proc, stmt_loop, loop, unsafe_disable_check=unsafe_disable_check)
    proc = repeate(reorder_stmt_forward)(proc, stmt)

    if not rc:
        return proc
    stmt = proc.forward(stmt)
    stmt_loop = proc.forward(stmt_loop)
    return proc, jam_stmt_cursors(stmt_loop, stmt)


def parallelize_reduction(
    proc, reduce_stmt, factor=None, memory=DRAM, nth_loop=1, unroll=False
):
    # Auto-coersion
    if isinstance(unroll, bool):
        unroll = (unroll, unroll)

    reduce_stmt = proc.forward(reduce_stmt)

    if not is_reduce(proc, reduce_stmt):
        raise TypeError(f"reduce_stmt must of type {ReduceCursor}")

    reduce_loop = get_enclosing_loop(proc, reduce_stmt, nth_loop)
    control_vars = get_symbols(proc, reduce_stmt.idx())
    if reduce_loop.name() in control_vars:
        raise SchedulingError("Statement isn't a reduction over the specified loop.")

    if factor is not None:
        proc, (reduce_loop, _, _) = divide_loop_(
            proc, reduce_loop, factor, tail="guard", rc=True
        )
        proc = simplify(proc)
        nth_loop += 1

    # Stage reduction around the loop we are reducing over
    proc = reorder_loops(proc, reduce_loop)
    proc = auto_stage_mem(
        proc,
        reduce_loop,
        reduce_stmt.name(),
        accum=True,
    )
    proc = simplify(proc)

    # Set the memory of the newly created buffer
    reduce_loop = proc.forward(reduce_loop)
    alloc = reduce_loop.prev().prev()
    proc = set_memory(proc, alloc, memory)

    # Parallelize the reduction
    reduce_loop = proc.forward(reduce_loop)
    proc = parallelize_and_lift_alloc(proc, reduce_loop.prev().prev())

    # Fission the zero and store-back stages
    proc = fission(proc, reduce_loop.before())
    proc = fission(proc, reduce_loop.after())

    # Reorder the loop nest back
    reduce_loop = proc.forward(reduce_loop)
    proc = reorder_loops(proc, reduce_loop.parent())

    # Unroll any loops
    reduce_loop = proc.forward(reduce_loop)
    if unroll[0]:
        proc = unroll_loop(proc, reduce_loop.prev())
    if unroll[1]:
        proc = unroll_loop(proc, reduce_loop.next())

    if factor is not None:
        proc = undo_divide_and_guard_loop(proc, reduce_loop)
    return proc


def parallelize_all_reductions(proc, loop, factor=None, memory=DRAM, unroll=False):
    loop = proc.forward(loop)

    def rewrite(proc, s):
        s = proc.forward(s)
        reduc_loop = proc.forward(loop)
        nth_loop = 0
        for parent in get_parents(proc, s):
            if is_loop(proc, parent):
                nth_loop += 1
            if parent == reduc_loop:
                break
        return parallelize_reduction(proc, s, factor, memory, nth_loop, unroll)

    return make_pass(attempt(rewrite))(proc, loop.body())


def unroll_and_jam(proc, loop, factor, unroll=(True, True, True)):
    loop = proc.forward(loop)
    inner_loops = [i for i in loop.body() if isinstance(i, ForCursor)]
    if len(inner_loops) > 1:
        raise SchedulingError("Multiple loops found, decision is ambigious")
    if len(inner_loops) == 0:
        raise SchedulingError("No loops found")

    return interleave_outer_loop_with_inner_loop(
        proc, loop, inner_loops[0], factor, unroll=unroll
    )


def unroll_and_jam_parent(proc, loop, factor, unroll=(True, True, True)):
    loop = proc.forward(loop)
    outer_loop = loop.parent()
    if not isinstance(outer_loop, ForCursor):
        raise SchedulingError("parent is not a loop")
    return interleave_outer_loop_with_inner_loop(
        proc, outer_loop, loop, factor, unroll=unroll
    )


def interleave_outer_loop_with_inner_loop(
    proc,
    outer_loop_cursor,
    inner_loop_cursor,
    interleave_factor,
    unroll=(True, True, True),
):
    # TODO: check if inner_loop is directly in the body of outer_loop
    outer_loop_cursor = proc.forward(outer_loop_cursor)
    inner_loop_cursor = proc.forward(inner_loop_cursor)

    if (
        isinstance(outer_loop_cursor.hi(), LiteralCursor)
        and outer_loop_cursor.hi().value() == interleave_factor
    ):
        middle_loop_cursor = outer_loop_cursor
    else:
        proc = divide_loop(
            proc,
            outer_loop_cursor,
            interleave_factor,
            (outer_loop_cursor.name() + "o", outer_loop_cursor.name() + "i"),
            tail="cut",
        )

        outer_loop_cursor = proc.forward(outer_loop_cursor)
        middle_loop_cursor = outer_loop_cursor.body()[0]
    middle_loop_stmts = list(middle_loop_cursor.body())

    proc = simplify(proc)
    for stmt in middle_loop_stmts:
        if isinstance(stmt, AllocCursor):
            proc = parallelize_and_lift_alloc(proc, stmt)
    inner_loop_cursor = proc.forward(inner_loop_cursor)

    if not isinstance(inner_loop_cursor.prev(), InvalidCursor):
        proc = fission(proc, inner_loop_cursor.before())
        inner_loop_cursor = proc.forward(inner_loop_cursor)
        if unroll[0]:
            proc = unroll_loop(proc, inner_loop_cursor.parent().prev())

    if not isinstance(inner_loop_cursor.next(), InvalidCursor):
        proc = fission(proc, inner_loop_cursor.after())
        inner_loop_cursor = proc.forward(inner_loop_cursor)
        if unroll[2]:
            proc = unroll_loop(proc, inner_loop_cursor.parent().next())

    inner_loop_cursor = proc.forward(inner_loop_cursor)

    proc = reorder_loops(proc, inner_loop_cursor.parent())
    if unroll[1]:
        proc = unroll_loop(proc, inner_loop_cursor.parent())

    return proc


def fission_into_singles(proc, cursor):
    if not isinstance(cursor, (ForCursor, IfCursor)):
        raise SchedulingError(
            f"Got type {type(cursor)}, expected {ForCursor} or {IfCursor}"
        )

    cursor = proc.forward(cursor)

    def dfs(proc, cursor, n_lifts=0):
        if n_lifts and not is_end_of_body(proc, cursor):
            proc = fission(proc, cursor.after(), n_lifts)
        children = get_children(proc, cursor)
        children = filter(lambda s: isinstance(s, StmtCursor), children)
        return apply(dfs)(proc, children, n_lifts + 1)

    proc = parallelize_allocs(proc, cursor)
    return dfs(proc, cursor)


@dataclass
class vectorize_cursors:
    loop: ForCursor

    def __iter__(self):
        yield self.loop


def vectorize_predicate_tail(
    proc,
    loop,
    vec_width,
    precision,
    mem_type,
    instructions=[],
    rules=[],
    tail="cut_and_predicate",
    rc=False,
):
    proc = parallelize_all_reductions(proc, loop, factor=vec_width, memory=mem_type)
    proc = stage_compute(proc, loop, precision, mem_type, rules)
    proc, (outer, inner, _) = divide_loop_(proc, loop, vec_width, rc=True)
    proc = simplify(proc)
    proc = fission_into_singles(proc, inner)

    if tail == "cut_and_predicate":
        if_c = inner.body()[0]
        cut = FormattedExprStr("_ - 1", outer.hi())
        proc = attempt(cut_loop)(proc, outer, cut)
        outer = proc.forward(outer)
        proc = dce(proc, outer.body())
    proc = replace_all_stmts(proc, instructions)

    if not rc:
        return proc
    outer = proc.forward(outer)
    return proc, vectorize_cursors(outer)


def vectorize(
    proc,
    loop,
    vec_width,
    precision,
    mem_type,
    instructions=[],
    rules=[],
    tail="cut_and_predicate",
    rc=False,
):
    if tail in {"predicate", "cut_and_predicate"}:
        return vectorize_predicate_tail(
            proc, loop, vec_width, precision, mem_type, instructions, rules, tail, rc
        )
    proc, (outer, inner, _) = divide_loop_(proc, loop, vec_width, tail=tail, rc=True)
    proc = parallelize_all_reductions(proc, outer, memory=mem_type)

    outer = proc.forward(outer)
    inner = outer.body()[0]

    proc = stage_compute(proc, inner, precision, mem_type, rules)
    proc = fission_into_singles(proc, inner)

    proc = replace_all_stmts(proc, instructions)

    if not rc:
        return proc
    outer = proc.forward(outer)
    return proc, vectorize_cursors(outer)


def tile_loops(proc, loop_tile_pairs, perfect=False):

    loop_tile_pairs = [(proc.forward(i[0]), i[1]) for i in loop_tile_pairs]

    inner_loops = []
    for i in range(len(loop_tile_pairs)):
        outer_loop = loop_tile_pairs[i][0]
        tile_size = loop_tile_pairs[i][1]
        new_names = (outer_loop.name() + "o", outer_loop.name() + "i")
        if perfect or (
            isinstance(outer_loop.hi(), LiteralCursor)
            and (outer_loop.hi().value() % tile_size == 0)
        ):
            proc = divide_loop(proc, outer_loop, tile_size, new_names, perfect=True)
        else:
            proc = divide_loop(proc, outer_loop, tile_size, new_names, tail="cut")
        inner_loop = proc.forward(outer_loop).body()[0]
        inner_loops.append(inner_loop)

    for i in range(len(loop_tile_pairs) - 2, -1, -1):
        inner_loop = inner_loops[i]
        tile_size = loop_tile_pairs[i][1]
        for j in range(i + 1, len(loop_tile_pairs)):
            loop = loop_tile_pairs[j][0]
            proc = interleave_outer_loop_with_inner_loop(
                proc, inner_loop, loop, tile_size, (False, False, False)
            )
    return proc, [proc.forward(l) for l in inner_loops]


def tile_loops_bottom_up(proc, loop, tiles):

    cur_loop = loop
    for i in tiles[:-1]:
        if not len(cur_loop.body()) == 1:
            raise SchedulingError("All loop must have a body length of 1")
        if not isinstance(cur_loop.body()[0], ForCursor):
            raise SchedulingError("Did not find a nested loop")
        cur_loop = cur_loop.body()[0]

    loops = []
    cur_loop = loop
    for i in tiles:
        loops.append((cur_loop, i))
        cur_loop = cur_loop.body()[0]

    def get_depth(loop):
        if not isinstance(loop, (ForCursor, IfCursor)):
            return 0
        return max([get_depth(i) for i in loop.body()]) + 1

    def push_loop_in(proc, loop, depth):
        loop = proc.forward(loop)
        if get_depth(loop) == depth:
            return proc
        count = len(loop.body())
        for stmt in list(loop.body())[:-1]:
            proc = fission(proc, stmt.after())
        loop = proc.forward(loop)
        loops = []
        for i in range(count):
            loops.append(loop)
            loop = loop.next()
        for loop in loops:
            if get_depth(loop) == depth:
                continue
            loop = proc.forward(loop)
            child = loop.body()[0]
            if isinstance(child, ForCursor):
                proc = reorder_loops(proc, loop)
                forwarded_loop = proc.forward(loop)
            elif isinstance(child, IfCursor):
                proc = lift_scope(proc, child)
                child = proc.forward(child)
                forwarded_loop = child.body()[0]
            else:
                assert False, "Invalid"
            proc = push_loop_in(proc, forwarded_loop, depth)
        return proc

    for depth, (loop, tile) in enumerate(loops[::-1]):
        if tile is not None:
            proc, (_, inner, tail) = divide_loop_(
                proc, loop, tile, tail="cut_and_guard", rc=True
            )
            proc = push_loop_in(proc, inner, depth + 1)
            proc = push_loop_in(proc, tail.body()[0], depth + 1)
        else:
            proc = push_loop_in(proc, loop, depth + 1)

    return proc


def auto_stage_mem(proc, block, buff, new_buff_name=None, accum=False, rc=False):
    if new_buff_name is None:
        new_buff_name = next(get_unique_names(proc))

    if not isinstance(block, BlockCursor):
        block = proc.forward(block)
        block = block.as_block()

    block_nodes = list(lrn(proc, block))
    block_loops = list(filter(lambda s: is_loop(proc, s), block_nodes))
    block_accesses = filter(
        lambda s: is_access(proc, s) and s.name() == buff, block_nodes
    )

    def eval_rng(expr, env):
        expr = proc.forward(expr)
        if is_binop(proc, expr):
            lhs_rng = eval_rng(expr.lhs(), env)
            rhs_rng = eval_rng(expr.rhs(), env)
            join = lambda l, op, r: f"({l}) {op} ({r})"
            rng_l = join(lhs_rng[0], expr.op(), rhs_rng[0])
            rng_r = join(lhs_rng[1], expr.op(), rhs_rng[1])
            if is_add(proc, expr) or is_sub(proc, expr):
                return (rng_l, rng_r)
            else:
                rng = (rng_l, rng_r)
                if is_literal(proc, expr.rhs()):
                    if expr.rhs().value() < 0:
                        rng = rng[::-1]
                elif is_literal(proc, expr.lhs()):
                    if expr.lhs().value() < 0:
                        rng = rng[::-1]
                else:
                    assert False, "Unreachable case"
                return rng
        elif is_unary_minus(proc, expr):
            rng = eval_rng(expr.arg(), env)
            return (f"-({rng[1]})", f"-({rng[0]})")
        elif is_read(proc, expr):
            if expr.name() in env:
                return env[expr.name()]
            else:
                read = f"{expr.name()}"
                return (read, read)
        elif is_literal(proc, expr):
            val = f"{expr.value()}"
            return (val, val)
        else:
            assert False, "Unreachable case"

    def get_window(cursor):
        parents = get_parents(proc, cursor)
        my_loops = list(filter(lambda s: s in block_loops, parents))
        my_loops = my_loops[::-1]
        env = {}
        for loop in my_loops:
            lo_rng = eval_rng(loop.lo(), env)
            hi_rng = eval_rng(loop.hi(), env)
            env[loop.name()] = (lo_rng[0], hi_rng[1])
        window = tuple(eval_rng(idx, env) for idx in cursor.idx())
        return window

    def window_to_str(window):
        def get_dim_win(dim):
            if dim[0] == dim[1]:
                return dim[0]
            return f"{dim[0]}:{dim[1]}"

        dims = [get_dim_win(i) for i in window]
        window = ",".join(dims)
        window = f"{buff}[{window}]" if window else buff
        return window

    def my_stage_mem(proc, window):
        window = window_to_str(window)
        return attempt(stage_mem, errs=(SchedulingError,))(
            proc, block, window, new_buff_name, accum, rs=True
        )

    declaration = get_declaration(proc, block[0], buff)
    if not declaration.is_tensor():
        return stage_mem(proc, block, buff, new_buff_name, accum)

    # Start with the window as the entire buffer
    buff_window = [["0", eval_rng(dim, {})[0]] for dim in declaration.shape()]

    # Get the window per access
    accessess_windows = [get_window(access) for access in block_accesses]

    # Prune the candidates for each dim x side
    candidates = [(set(), set()) for i in buff_window]
    for access_window in accessess_windows:
        for i, dim in enumerate(access_window):
            for side in range(2):
                candidates[i][side].add(dim[side])
                candidates[i][side].add(dim[side])

    # Tighten each dimension and side of the buffer window
    for i, dim in enumerate(buff_window):
        for side in range(0, 2):
            # Prune
            if len(candidates[i][side]) == 1:
                dim[side] = next(iter(candidates[i][side]))
                continue
            for candidate in candidates[i][side]:
                old_side = dim[side]
                dim[side] = candidate
                new_proc, success = my_stage_mem(proc, buff_window)
                if not success:
                    dim[side] = old_side
    window = window_to_str(buff_window)
    return stage_mem_(proc, block, window, new_buff_name, accum, rc=rc)


def ordered_stage_expr(proc, expr_cursors, new_buff_name, precision, n_lifts=1):
    if not isinstance(expr_cursors, list):
        expr_cursors = [expr_cursors]

    if not all([isinstance(cursor, ExprCursor) for cursor in expr_cursors]):
        raise SchedulingError("auto_stage_mem expects a read a cursor")

    expr_cursors = [proc.forward(c) for c in expr_cursors]
    original_stmt = get_enclosing_stmt(proc, expr_cursors[0])

    proc = bind_expr(proc, expr_cursors, new_buff_name)
    original_stmt = proc.forward(original_stmt)
    assign_cursor = original_stmt.prev()
    alloc_cursor = assign_cursor.prev()
    expr_cursor = assign_cursor.rhs()
    deps = list(get_symbols(proc, expr_cursor))

    assert isinstance(assign_cursor, AssignCursor)
    assert isinstance(alloc_cursor, AllocCursor)

    anchor_stmt = assign_cursor

    def hoist_as_loop(proc, stmt_cursor):
        stmt_cursor = proc.forward(stmt_cursor)
        while not isinstance(stmt_cursor.prev(), InvalidCursor):
            proc = reorder_stmts(proc, stmt_cursor.expand(1, 0))
            stmt_cursor = proc.forward(stmt_cursor)

        proc = fission(proc, stmt_cursor.after())

        return proc

    for i in range(n_lifts):
        parent = anchor_stmt.parent()

        if not isinstance(parent, ForCursor):
            raise SchedulingError("Not implemented yet")
        if parent.name() in deps:
            proc = parallelize_and_lift_alloc(proc, alloc_cursor)
        else:
            proc = lift_alloc(proc, alloc_cursor)

        proc = hoist_as_loop(proc, anchor_stmt)
        anchor_stmt = proc.forward(anchor_stmt)
        anchor_stmt = anchor_stmt.parent()

    alloc_cursor = proc.forward(alloc_cursor)
    loop_nest = alloc_cursor.next()

    def try_removing_loops(proc, loop):
        child_stmt = loop.body()[0]
        if isinstance(child_stmt, ForCursor):
            proc = try_removing_loops(proc, child_stmt)
        try:
            proc = remove_loop(proc, loop)
        except:
            pass
        return proc

    proc = try_removing_loops(proc, loop_nest)
    alloc_cursor = proc.forward(alloc_cursor)
    proc = set_precision(proc, alloc_cursor, precision)
    scopes_nest = alloc_cursor.next()

    def lift_all_ifs(proc, scope, depth=0):
        if isinstance(scope, IfCursor):
            for i in range(depth):
                proc = lift_scope(proc, scope)
        child_stmt = scope.body()[0]
        if isinstance(child_stmt, (ForCursor, IfCursor)):
            proc = lift_all_ifs(proc, child_stmt, depth + 1)
        return proc

    proc = lift_all_ifs(proc, scopes_nest)

    return proc


def _eliminate_dead_code_pruned(proc, s):
    s = proc.forward(s)
    if isinstance(s, ForCursor) and is_loop_bounds_const(s):
        return proc
    else:
        return eliminate_dead_code(proc, s)


dce = make_pass(attempt(_eliminate_dead_code_pruned))


def unroll_buffers(proc, block=InvalidCursor(), mem=None):
    def rewrite(proc, alloc):
        alloc = proc.forward(alloc)
        if not isinstance(alloc, AllocCursor):
            return proc
        if not alloc.is_tensor():
            return proc
        diff = int(alloc.mem() is mem)
        if len(alloc.shape()) - diff:
            if isinstance(alloc.shape()[0], LiteralCursor):
                return unroll_buffer(proc, alloc, 0)
        return proc

    while True:
        new_proc = make_pass(rewrite)(proc, block)
        if new_proc == proc:
            break
        proc = new_proc
    return new_proc


def unfold_reduce(proc, reduce):
    if not isinstance(reduce, ReduceCursor):
        raise SchedulingError("Expected a reduce cursor")

    proc = auto_stage_mem(proc, reduce, reduce.name())
    reduce = proc.forward(reduce)
    alloc = reduce.prev().prev()
    proc = merge_writes(proc, reduce.as_block().expand(delta_lo=1, delta_hi=0))
    assign = proc.forward(alloc).next()
    proc = inline_assign(proc, assign)
    proc = delete_buffer(proc, alloc)

    return proc


def fma_rule(proc, expr):
    expr = proc.forward(expr)

    if is_add(proc, expr):
        if is_mul(proc, expr.lhs()):
            # (a * b) + c
            return [expr.lhs().lhs(), expr.lhs().rhs(), expr.rhs()]
        elif is_mul(proc, expr.rhs()):
            # a + (b * c)
            return [expr.lhs(), expr.rhs().lhs(), expr.rhs().rhs()]

    return None


def abs_rule(proc, expr):
    expr = proc.forward(expr)
    if is_select(proc, expr):
        args = expr.args()
        if (
            is_literal(proc, args[0], 0.0)
            and is_unary_minus(proc, args[3])
            and are_exprs_equal(proc, args[1], args[2])
            and are_exprs_equal(proc, args[1], args[3].arg())
        ):
            return [[args[1], args[2], args[3].arg()]]
    return None


def stage_expr_into_memory(proc, exprs, precision, memory):
    if not isinstance(exprs, list):
        exprs = [exprs]

    expr = proc.forward(exprs[0])

    # No need to stage if expr is already assigned
    # to the target memory
    parent = expr.parent()
    if (
        isinstance(parent, AssignCursor)
        and get_declaration(proc, expr, parent.name()).mem() is memory
    ):
        return proc, expr

    # No need to stage if expr is already read
    # from the target memory
    if (
        isinstance(expr, ReadCursor)
        and get_declaration(proc, expr, expr.name()).mem() is memory
    ):
        return proc, expr
    return lift_rc(bind_and_set_expr, "bound_expr")(proc, exprs, precision, memory)


def stage_compute(
    proc,
    block=InvalidCursor(),
    precision="R",
    memory=DRAM,
    children_ops=[],
):

    if not isinstance(children_ops, list):
        raise SchedulingError("Expected children_ops to be a list")

    def get_numeric_children(proc, cursor=InvalidCursor()):
        check = lambda c: hasattr(c, "type") and c.type().is_numeric()
        yield from filter(check, get_children(proc, cursor))

    children_ops.append(get_numeric_children)

    def stage(proc, exprs):
        proc, expr = stage_expr_into_memory(proc, exprs, precision, memory)
        for children_op in children_ops:
            if children := children_op(proc, expr):
                break
        return apply(stage)(proc, children)

    allocs = filter(lambda s: isinstance(s, AllocCursor), nlr_stmts(proc, block))
    proc = apply(set_memory)(proc, allocs, memory)
    proc = make_pass(attempt(unfold_reduce))(proc, block)
    assigns = filter(lambda s: isinstance(s, AssignCursor), lrn_stmts(proc, block))
    exprs = [assign.rhs() for assign in assigns]
    proc = apply(stage)(proc, exprs)
    # TODO: uncomment once bug in Exo is fixed
    # proc = inline_copies(proc, block)
    proc = make_pass(attempt(fold_into_reduce))(proc, block)
    proc = dealiasing_pass(proc, block)
    return proc


def check_call_site(proc, call_cursor):
    if not isinstance(call_cursor, CallCursor):
        raise TypeError("call_cursor must be a CallCursor")

    ###################################################################
    # build an env of symbols this call statement observes.
    # e.g. {x: (DRAM, "f32"), y: (Neon, "f64")}
    ###################################################################
    env = {}
    obs_stmts = get_observed_stmts(call_cursor)
    allocs = filter(lambda s: isinstance(s, AllocCursor), obs_stmts)
    for s in list(proc.args()) + list(allocs):
        if s.type().is_numeric():
            mem = s.mem()
            env[s.name()] = (mem, s.type())
    ###################################################################
    # Check consistency at call site
    ###################################################################
    call_args = call_cursor.args()
    callee_parameters = call_cursor.subproc().args()
    for arg, par in zip(call_args, callee_parameters):
        par_type = par.type()
        if par_type.is_numeric():
            par_mem = par.mem()
            arg_mem, arg_type = env[arg.name()]
            if not issubclass(arg_mem, par_mem) or arg_type is not par_type:
                return False
    return True


@dataclass
class replace_cursors:
    call: CallCursor

    def __iter__(self):
        yield self.call


# TODO: change this to work for blocks
def checked_replace(proc, stmt, subproc, quiet=False):
    stmt = proc.forward(stmt)
    parent = stmt.parent()
    index = get_index_in_body(proc, stmt)

    try:
        proc = replace(proc, stmt, subproc, quiet=quiet)
    except:
        raise SchedulingError("failed to replace")

    is_else = False
    if (
        isinstance(parent, IfCursor)
        and not isinstance(parent.orelse(), InvalidCursor)
        and index < len(parent.orelse())
        and parent.orelse()[index] == stmt
    ):
        is_else = True
    if not isinstance(parent, InvalidCursor):
        parent = proc.forward(parent)
    else:
        parent = proc

    call = parent.body()[index] if not is_else else parent.orelse()[index]
    if not check_call_site(proc, call):
        raise SchedulingError("Call site inconsistency")
    return proc


def replace_all_stmts(proc, instructions):
    if not isinstance(instructions, list):
        instructions = [instructions]

    for stmt in nlr_stmts(proc):
        try:
            stmt = proc.forward(stmt)
        except InvalidCursorError:
            continue

        for instr in instructions:
            try:
                proc = checked_replace(proc, stmt, instr, quiet=True)
                break
            except SchedulingError:
                pass
    return proc


def bound_loop_by_if(proc, loop):
    loop = proc.forward(loop)
    err = "Expected loop to be of the following structure:\nfor iter in seq(lo, hi):\n\t if iter < e:"
    if len(loop.body()) != 1 or not isinstance(loop.body()[0], IfCursor):
        raise SchedulingError(err)

    if_c = loop.body()[0]
    if not isinstance(if_c.orelse(), InvalidCursor):
        raise SchedulingError(err)

    if (
        not isinstance(if_c.cond().lhs(), ReadCursor)
        or if_c.cond().lhs().name() != loop.name()
        or if_c.cond().op() != "<"
    ):
        raise SchedulingError(err)

    if_c = loop.body()[0]
    proc = cut_loop(proc, loop, FormattedExprStr("_ + _", loop.lo(), if_c.cond().rhs()))
    loop1 = proc.forward(loop)
    loop2 = loop1.next()
    proc = eliminate_dead_code(proc, loop1.body()[0])
    # This should step can be skipped in general, but there is a problem in
    # Exo where the check of the inner if would fail if the loop bounds
    # are equal or dead in general.
    proc, success = attempt(eliminate_dead_code)(proc, loop2, rs=True)
    if success:
        return proc
    proc = eliminate_dead_code(proc, loop2.body()[0])
    # proc = delete_pass(proc, loop2), but it doesn't forward it
    return proc


def undo_divide_and_guard_loop(proc, loop):
    loop = proc.forward(loop)
    proc = mult_loops(proc, loop, loop.name()[:-1])
    proc = simplify(proc)
    proc = bound_loop_by_if(proc, loop)
    return proc


def unroll_loops(proc, block=InvalidCursor(), threshold=None):
    def pred(proc, s):
        s = proc.forward(s)
        if not (is_loop(proc, s) and is_loop_bounds_const(proc, s)):
            return False
        return threshold is None or s.hi().value() - s.lo().value() <= threshold

    return make_pass(predicate(unroll_loop, pred), lrn_stmts)(proc, block)


def cleanup(proc):
    proc = simplify(proc)
    proc = unroll_loops(proc, threshold=1)
    proc = dce(proc)
    try:
        proc.find("pass")
        proc = delete_pass(proc)
    except SchedulingError:
        pass
    return proc


def reorder_stmt_forward(proc, stmt):
    stmt = proc.forward(stmt)
    block = stmt.as_block().expand(0, 1)
    return reorder_stmts(proc, block)


def reorder_stmt_backwards(proc, stmt):
    stmt = proc.forward(stmt)
    block = stmt.as_block().expand(-1, 0)
    return reorder_stmts(proc, block)


@dataclass
class divide_loop_recursive_cursors:
    outer_loops: list
    inner_loops: list
    tail_loop: ForCursor

    def __iter__(self):
        yield self.outer_loops
        yield self.inner_loops
        yield self.tail_loop


def divide_loop_recursive(proc, loop, factor, tail="cut", rc=False):
    if tail not in {"cut", "cut_and_guard"}:
        raise SchedulingError("tail strategy must be cut or cut_and_guard")
    outer_loops = []
    inner_loops = []
    tail_loop = loop
    while factor > 1:
        proc, (outer, inner, tail_loop) = divide_loop_(
            proc, tail_loop, factor, tail=tail, rc=True
        )
        outer_loops.append(outer)
        inner_loops.append(inner)
        factor = factor // 2
    if not rc:
        return proc
    outer_loops = [proc.forward(c) for c in outer_loops]
    inner_loops = [proc.forward(c) for c in inner_loops]
    return proc, divide_loop_recursive_cursors(outer_loops, inner_loops, tail_loop)


@dataclass
class binary_specialize_cursors:
    stmts: Cursor

    def __iter__(self):
        yield self.stmts


def binary_specialize(proc, stmt, expr, values, rc=False):
    stmt = proc.forward(stmt)
    if isinstance(expr, ExprCursor):
        expr = proc.forward(expr)
        expr = expr_to_string(expr)
    get_cond = lambda op, v: f"{expr} {op} {v}"

    if len(values) == 1:
        raise SchedulingError("Cannot specialize given one value!")
    values = sorted(values)
    stmt = proc.forward(stmt)

    stmts = []

    def rewrite(proc, stmt, values):
        if len(values) == 1:
            # This should be redundant if the user provided correct inputs!
            # So, it is really a check that the inputs the user provided cover the full range.
            proc, (if_stmt,) = specialize_(
                proc, stmt, get_cond("==", values[0]), rc=True
            )
            proc = simplify(proc)
            proc = eliminate_dead_code(proc, if_stmt)
            stmts.append(if_stmt.body()[0])
            stmts.append(if_stmt.orelse()[0])
            return proc
        md = len(values) // 2
        proc, (if_stmt,) = specialize_(proc, stmt, get_cond("<", values[md]), rc=True)
        proc = rewrite(proc, if_stmt.body()[0], values[:md])
        proc = rewrite(proc, if_stmt.orelse()[0], values[md:])
        return proc

    proc = rewrite(proc, stmt, values)
    if not rc:
        return proc

    filtered_stmts = []
    for s in stmts:
        try:
            stmt = proc.forward(s)
            if not isinstance(stmt, PassCursor):
                filtered_stmts.append(stmt)
        except InvalidCursorError:
            pass
    return proc, binary_specialize_cursors(filtered_stmts)


def cse(proc, block, precision):
    if not isinstance(block, BlockCursor):
        block = proc.forward(block).as_block()

    nodes = list(lrn(proc, block))

    # First do CSE on the buffer accesses
    accesses = filter(lambda c: is_access(proc, c), nodes)
    accesses = filter(lambda c: is_stmt(proc, c) or c.type().is_numeric(), accesses)

    buff_map = {}

    for access in accesses:
        idx_ac = buff_map.setdefault(access.name(), {})
        idx_str = expr_to_string(access.idx())
        idx_ac.setdefault(idx_str, []).append(access)

    for buff, idx_map in buff_map.items():
        for idx, access_list in idx_map.items():
            if len(access_list) > 1:
                if all(is_read(proc, c) for c in access_list):
                    proc = bind_and_set_expr(proc, access_list, precision, DRAM)
                else:
                    staging_block = get_bounding_block(proc, access_list)
                    proc = auto_stage_mem(proc, staging_block, buff)
    return proc


inline_copies = make_pass(predicate(attempt(inline_assign), is_copy))


def dealias(proc, stmt):
    stmt = proc.forward(stmt)

    if not is_assign(proc, stmt) and not is_reduce(proc, stmt):
        raise TypeError("stmt must be an assign or reduce")

    nodes = list(lrn(proc, stmt.rhs()))

    accesses = filter(lambda c: is_access(proc, c), nodes)
    accesses = filter(lambda c: is_stmt(proc, c) or c.type().is_numeric(), accesses)

    buff_map = {}
    for access in accesses:
        buff_map.setdefault(access.name(), []).append(access)

    for buff, buff_accesses in buff_map.items():
        for access in buff_accesses[:-1]:
            decl = get_declaration(proc, stmt, access.name())
            proc = bind_and_set_expr(proc, access, decl.type(), decl.mem())

    if stmt.name() in buff_map:
        decl = get_declaration(proc, stmt, stmt.name())
        proc = bind_and_set_expr(proc, stmt.rhs(), decl.type(), decl.mem())

    return proc


dealiasing_pass = make_pass(attempt(dealias, errs=(TypeError,)))


def round_loop(proc, loop, factor, up=True):
    tail = "guard" if up else "cut"
    proc, (loop, _, _) = divide_loop_(proc, loop, factor, tail=tail, rc=True)
    proc = mult_loops(proc, loop, loop.name()[:-1])
    return proc


@dataclass
class cut_loop_and_unroll_cursors:
    loop: ForCursor

    def __iter__(self):
        yield self.loop


def cut_loop_and_unroll(proc, loop, const, front=True, rc=False):
    loop = proc.forward(loop)
    cut = (
        FormattedExprStr("_ + 1", loop.lo())
        if front
        else FormattedExprStr("_ - 1", loop.hi())
    )
    proc, (const_loop, loop) = cut_loop_(proc, loop, cut, rc=True)
    if not front:
        const_loop, loop = loop, const_loop
    proc = shift_loop(proc, const_loop, 0)
    proc = simplify(proc)
    proc = unroll_loop(proc, const_loop)
    proc = shift_loop(proc, loop, 0)
    if not rc:
        return proc
    return proc, cut_loop_and_unroll_cursors(loop)



# import inspect
# import types
from weakref import WeakKeyDictionary

from collections import defaultdict  # ChainMap, OrderedDict
from itertools import chain
from dataclasses import dataclass
from typing import Any, Optional
from enum import Enum

from .LoopIR import LoopIR, T, Alpha_Rename

# This file keeps track of equivalence of procedures due to
# scheduling.  We sometimes refer to this as "provenance" since
# equality is tracked based on the transitive chain of equivalences
# generated through the act of scheduling

# A lot of this is taken from the Wikipedia article
# on disjoint-set data structures (i.e. union-find)

# However, the whole issue is complicated by the idea that we want to
# reason about equivalence "modulo a set of keys"
# (these keys are global/configuration variables)
# For instance, consider two procedures x and y.  We want to say
# that they're equivalent "modulo"/"except for" a set {a,b,c}
#       Write this x =={a,b,c} y.
#
# If L, K are sets s.t. L <= K (is a subset), then
#       x ==L y   ==>   x ==K y
#
# Furthermore, we want to observe a further property between our equivalences.
# If L, K are sets and L^K is their intersection,
#       (x ==L y /\ x ==K y)   ==>   x ==(L^K) y
#
# From these two properties, we can see that we have a semi-lattice defined
# on our entire space of different equivalence relations, where the
# equivalence corresponding to the emptyset is the strictest equality
# (i.e. the fewest things are said to be equal).
#
# If we could somehow enumerate a set of all the "weakest" equalities
# then we could track all of those equivalence relations and reconstruct
# an equivalence query for a particular set of keys by intersecting all
# of the appropriate "basis"/"atomic" relations.  Unfortunately the set
# of possible global variables is not bounded a priori.
#
# Despite this fact, we can track everything by hypothesizing a "universe"
# of all keys called `Unv`.  Then at any point in time where this code module
# has been made aware of keys {x1, x2, ..., xn} we can track n+1 equivalence
# relations, corresponding to
#   Unv, Unv-{x1}, Unv-{x2}, ..., Unv-{xn}
# If we add a new global key `y`, then we can form `Unv-{y}` to begin with
# by simply copying the equivalence relation `Unv`, which must be the same
# as `Unv-{y}` until we start observing equivalences modulo `y` (at which
# point those equivalences WILL be added to `Unv` but not to `Unv-{y}`)
#
# We can always recover the strictest equality `x =={} y` by intersecting
# all currently tracked equivalence relations.  Operationally, this can
# be achieved by checking that an equivalence is in all tracked equivalences
#

# This whole scheme will almost certainly scale empirically as
#  O(a(#P) * #G) where #P is the number of procedures being tracked
# for equivalence, and #G is the number of global variables being tracked
# for the sake of equivalence modulo that global.
# Because #G does not get very large and because a(#P) is
# significantly sublinear, we expect that this scheme will suffice
# for a good long while.  (Although note that memory consumption
# is O(#P * #G))


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
#   basic union find datastructure


class _UnionFind:
    def __init__(self):
        self.lookup = WeakKeyDictionary()

    def new_node(self, val):
        if val not in self.lookup:
            self.lookup[val] = val

    def find(self, val):
        parent = self.lookup[val]
        while val is not parent:
            # path splitting optimization
            grandparent = self.lookup[parent]
            self.lookup[val] = grandparent
            val, parent = parent, grandparent
        return val

    def union(self, val1, val2):
        p1, p2 = self.find(val1), self.find(val2)

        if p1 is p2:
            pass  # then val1 and val2 are already unified
        else:
            self.lookup[p2] = p1

    def check_eqv(self, val1, val2):
        p1, p2 = self.find(val1), self.find(val2)
        return p1 is p2

    def copy_entire_UF(self):
        copy = _UnionFind()
        for v, p in self.lookup.items():
            copy.lookup[v] = p
        return copy


_UF_Unv = _UnionFind()
_UF_Strict = _UnionFind()
_UF_Unv_key = dict()


def new_uf_by_eqv_key(key):
    assert key not in _UF_Unv_key
    _UF_Unv_key[key] = _UF_Unv.copy_entire_UF()


def decl_new_proc(proc):
    # add to all existing union-find data structures
    _UF_Strict.new_node(proc)
    _UF_Unv.new_node(proc)
    for uf in _UF_Unv_key.values():
        uf.new_node(proc)


def derive_proc(orig_proc, new_proc, config_set=frozenset()):
    decl_new_proc(new_proc)
    assert_eqv_proc(orig_proc, new_proc, config_set)


def assert_eqv_proc(proc1, proc2, config_set=frozenset()):
    assert isinstance(config_set, frozenset)
    # First, expand the set of equivalences being tracked if needed
    for key in config_set:
        if key not in _UF_Unv_key:
            new_uf_by_eqv_key(key)
    # then do the appropriate union operations
    if not config_set:
        _UF_Strict.union(proc1, proc2)
    _UF_Unv.union(proc1, proc2)
    for key, uf in _UF_Unv_key.items():
        if key not in config_set:
            uf.union(proc1, proc2)


def check_eqv_proc(proc1, proc2, config_set=frozenset()):
    assert isinstance(config_set, frozenset)
    # if these aren't equal under the weakest assumptions
    # then we can early exit
    if not _UF_Unv.check_eqv(proc1, proc2):
        return False
    # otherwise check intersection of all non-excluded equivalence relations
    return all(
        uf.check_eqv(proc1, proc2)
        for key, uf in _UF_Unv_key.items()
        if key not in config_set
    )


def get_strictest_eqv_proc(proc1, proc2):
    # under the weakest assumptions, are these procedures equivalent?
    is_eqv = _UF_Unv.check_eqv(proc1, proc2)
    # then compute the strongest assumptions under which the equivalence
    # continues to hold.  Note that keys == emptyset() is the strictest
    keys = set()
    if is_eqv:
        keys = {
            key for key, uf in _UF_Unv_key.items() if not uf.check_eqv(proc1, proc2)
        }

    return is_eqv, keys


def get_repr_proc(q_proc):
    proc = _UF_Strict.find(q_proc)
    return proc



import functools
import itertools
import re
from collections import ChainMap

import pysmt
from asdl_adt import ADT
from pysmt import shortcuts as SMT

from .LoopIR import LoopIR, T, LoopIR_Do, FreeVars, Alpha_Rename, comparision_ops
from .LoopIR_dataflow import LoopIR_Dependencies
from .LoopIR_scheduling import SchedulingError
from .prelude import *
from .new_eff import Check_Aliasing
import exo.internal_cursors as ic


def _get_smt_solver():
    factory = pysmt.factory.Factory(pysmt.shortcuts.get_env())
    slvs = factory.all_solvers()
    if len(slvs) == 0:
        raise OSError("Could not find any SMT solvers")
    return pysmt.shortcuts.Solver(name=next(iter(slvs)))


def sanitize_str(s):
    return re.sub(r"\W", "_", s)


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #


class UnificationError(Exception):
    def __init__(self, msg):
        self._err_msg = str(msg)

    def __str__(self):
        return self._err_msg


def Get_Live_Variables(stmt_cursor):
    c = stmt_cursor
    covering_stmts = []
    while True:
        try:
            c = c.prev()
        except ic.InvalidCursorError:
            c = c.parent()
            if c == c.root():
                break
        covering_stmts.append(c._node)

    live_vars = ChainMap()

    proc = c.get_root()
    for arg in proc.args:
        live_vars[arg.name] = arg.type

    for s in reversed(covering_stmts):
        if isinstance(s, LoopIR.WindowStmt):
            live_vars[s.name] = s.rhs.type
        elif isinstance(s, LoopIR.Alloc):
            live_vars[s.name] = s.type
        elif isinstance(s, LoopIR.If):
            live_vars = live_vars.new_child()
        elif isinstance(s, LoopIR.For):
            live_vars = live_vars.new_child()
            live_vars[s.iter] = T.index

    return live_vars


def DoReplace(subproc, block_cursor):
    n_stmts = len(subproc.body)
    if len(block_cursor) < n_stmts:
        raise SchedulingError("Not enough statements to match")

    # prevent name clashes between the statement block and sub-proc
    temp_subproc = Alpha_Rename(subproc).result()
    stmts = [c._node for c in block_cursor[:n_stmts]]
    live_vars = Get_Live_Variables(block_cursor[0])
    new_args = Unification(temp_subproc, stmts, live_vars).result()

    # but don't use a different LoopIR.proc for the callsite itself
    new_call = LoopIR.Call(subproc, new_args, None, stmts[0].srcinfo)

    ir, fwd = block_cursor._replace([new_call])
    Check_Aliasing(ir)
    return ir, fwd


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Unification "System of Equations" grammar

# the problem designates disjoint subsets of the variables used
# as "holes" (variables to solve for) and "knowns" (variables to express
# a solution as an affine combination of).  Any variable not in either of
# those lists is unknown but not permissible in a solution expression.
UEq = ADT(
    """
module UEq {
    problem = ( sym*    holes,  -- symbols the solution is requested for
                sym*    knowns, -- symbols allowed in solution expressions
                pred*   preds   -- conj of equations
              )

    pred    = Conj( pred* preds )
            | Disj( pred* preds )
            | Cases( sym case_var, pred* cases )
            | Eq( expr lhs, expr rhs )

    -- affine expressions
    expr  =  Const(int val)
          |  Var( sym name )
          |  Add( expr lhs, expr rhs )
          |  Scale( int coeff, expr e )

} """,
    {
        "sym": Sym,
    },
)


# -------------------------------------- #
# Conversion to Strings for Debug


def _str_uexpr(e, prec=0):
    etyp = type(e)
    if etyp is UEq.Const:
        return str(e.val)
    elif etyp is UEq.Var:
        return str(e.name)
    elif etyp is UEq.Add:
        s = f"{_str_uexpr(e.lhs, 0)} + {_str_uexpr(e.rhs, 1)}"
        if prec > 0:
            s = f"({s})"
        return s
    elif etyp is UEq.Scale:
        return f"{e.coeff}*{_str_uexpr(e.e, 10)}"
    else:
        assert False, "bad case"


@extclass(UEq.Const)
@extclass(UEq.Var)
@extclass(UEq.Add)
@extclass(UEq.Scale)
def __str__(self):
    return _str_uexpr(self)


del __str__


def _str_upred(p, prec=0):
    ptyp = type(p)
    if ptyp is UEq.Eq:
        return f"{p.lhs} == {p.rhs}"
    elif ptyp is UEq.Conj or ptyp is UEq.Disj:
        op = " and " if UEq.Conj else " or "
        s = op.join([_str_upred(pp, 1) for pp in p.preds])
        if prec > 0:
            s = f"({s})"
        return s
    elif ptyp is UEq.Cases:
        return f"cases({p.case_var}) " + " | ".join([f"({pred})" for pred in p.cases])
    else:
        assert False, "bad case"


@extclass(UEq.Conj)
@extclass(UEq.Cases)
@extclass(UEq.Disj)
@extclass(UEq.Eq)
def __str__(self):
    return _str_upred(self)


del __str__


@extclass(UEq.problem)
def __str__(prob):
    lines = [
        "Holes:   " + ", ".join([str(x) for x in prob.holes]),
        "Knowns:  " + ", ".join([str(x) for x in prob.knowns]),
    ]
    lines += [str(p) for p in prob.preds]
    return "\n".join(lines)


del __str__


# -------------------------------------- #
# How to solve this system of equations


@extclass(UEq.expr)
def normalize(orig_e):
    def to_nform(e):
        if isinstance(e, UEq.Const):
            return [], e.val
        elif isinstance(e, UEq.Var):
            return [(1, e.name)], 0
        elif isinstance(e, UEq.Add):
            xcs, xoff = to_nform(e.lhs)
            ycs, yoff = to_nform(e.rhs)

            def merge(xcs=xcs, ycs=ycs):
                xi, yi = 0, 0
                res = []
                while xi < len(xcs) and yi < len(ycs):
                    (xc, xv), (yc, yv) = xcs[xi], ycs[yi]
                    if xv < yv:
                        res.append((xc, xv))
                        xi += 1
                    elif yv < xv:
                        res.append((yc, yv))
                        yi += 1
                    else:
                        if xc + yc != 0:
                            res.append((xc + yc, xv))
                        xi += 1
                        yi += 1
                if xi == len(xcs):
                    res += ycs[yi:]
                elif yi == len(ycs):
                    res += xcs[xi:]
                else:
                    assert False, "bad iteration"
                return res

            return merge(), xoff + yoff
        elif isinstance(e, UEq.Scale):
            if e.coeff == 0:
                return [], 0
            cs, off = to_nform(e.e)
            return [(e.coeff * c, v) for (c, v) in cs], e.coeff * off
        else:
            assert False, "bad case"

    def from_nform(cs, off):
        e = None
        for (c, v) in cs:
            assert c != 0
            t = UEq.Var(v)
            t = t if c == 1 else UEq.Scale(c, t)
            e = t if e is None else UEq.Add(e, t)
        if e is None:
            return UEq.Const(off)
        elif off == 0:
            return e
        else:
            return UEq.Add(e, UEq.Const(off))

    cs, off = to_nform(orig_e)
    return from_nform(cs, off)


@extclass(UEq.expr)
def sub(x, y):
    return UEq.Add(x, UEq.Scale(-1, y))


@extclass(UEq.problem)
def solve(prob):
    solver = _get_smt_solver()

    known_list = prob.knowns
    known_idx = {k: i for i, k in enumerate(known_list)}
    hole_idx = {k: i for i, k in enumerate(prob.holes)}
    Nk = len(known_list)

    var_set = dict()
    case_set = dict()

    def get_var(x):
        if x in var_set:
            return var_set[x]
        else:
            vec = [SMT.Symbol(f"{repr(x)}_{repr(k)}", SMT.INT) for k in known_list] + [
                SMT.Symbol(f"{repr(x)}_const", SMT.INT)
            ]
            var_set[x] = vec
            return vec

    def get_case(x):
        if x not in case_set:
            case_set[x] = SMT.Symbol(f"{repr(x)}", SMT.INT)
        return case_set[x]

    # initialize all hole variables, ensuring they are defined
    for x in prob.holes:
        get_var(x)

    def lower_e(e):
        if isinstance(e, UEq.Const):
            return ([SMT.Int(0)] * Nk) + [SMT.Int(e.val)]
        elif isinstance(e, UEq.Var):
            if e.name in known_idx:
                one_hot = [SMT.Int(0)] * (Nk + 1)
                one_hot[known_idx[e.name]] = SMT.Int(1)
                return one_hot
            elif e.name in hole_idx:
                return get_var(e.name)
            else:
                raise UnificationError(
                    f"Unable to cancel variable '{e.name}' from both sides "
                    f"of a unification equation"
                )
        elif isinstance(e, UEq.Add):
            lhs = lower_e(e.lhs)
            rhs = lower_e(e.rhs)
            return [SMT.Plus(x, y) for x, y in zip(lhs, rhs)]
        elif isinstance(e, UEq.Scale):
            arg = lower_e(e.e)
            return [SMT.Times(SMT.Int(e.coeff), a) for a in arg]
        else:
            assert False, "bad case"

    def lower_p(p):
        if isinstance(p, UEq.Eq):
            diff = p.lhs.sub(p.rhs).normalize()
            try:
                es = lower_e(diff)
                return SMT.And(*[SMT.Equals(x, SMT.Int(0)) for x in es])
            except UnificationError:
                return SMT.FALSE()
        elif isinstance(p, UEq.Conj):
            return SMT.And(*[lower_p(pp) for pp in p.preds])
        elif isinstance(p, UEq.Disj):
            return SMT.Or(*[lower_p(pp) for pp in p.preds])
        elif isinstance(p, UEq.Cases):
            case_var = get_case(p.case_var)

            def per_case(i, c):
                pp = lower_p(c)
                is_case = SMT.Equals(case_var, SMT.Int(i))
                return SMT.And(is_case, pp)

            disj = SMT.Or(*[per_case(i, c) for i, c in enumerate(p.cases)])
            case_lo = SMT.GE(case_var, SMT.Int(0))
            case_hi = SMT.LT(case_var, SMT.Int(len(p.cases)))
            return SMT.And(disj, case_lo, case_hi)
        else:
            assert False, "bad case"

    prob_pred = SMT.And(*[lower_p(p) for p in prob.preds])
    if not solver.is_sat(prob_pred):
        return None
    else:
        solutions = dict()
        for hole_var in prob.holes:
            x_syms = get_var(hole_var)
            x_val_dict = solver.get_py_values(x_syms)
            x_vals = [x_val_dict[x_sym] for x_sym in x_syms]
            expr = None
            for xx, v in zip(known_list, x_vals):
                v = int(v)
                if v == 0:
                    continue
                elif v == 1:
                    term = UEq.Var(xx)
                else:
                    term = UEq.Scale(v, UEq.Var(xx))

                expr = term if expr is None else UEq.Add(expr, term)

            # constant offset
            off = UEq.Const(int(x_vals[-1]))
            expr = off if expr is None else UEq.Add(expr, off)

            solutions[hole_var] = expr

        # report on case decisions
        for x in case_set:
            val = solver.get_py_value(case_set[x])
            solutions[x] = int(val)

        return solutions


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Unification compiler pass


class _Find_Mod_Div_Symbols(LoopIR_Do):
    def __init__(self, stmts, FV):
        self.node_to_sym = dict()  # many to one
        self.tuple_to_sym = dict()  # de-duplicating lookup
        self.sym_to_node = dict()  # pick a node for each symbol
        self.FV = FV

        self.unq_count = 0

        self.do_stmts(stmts)

    def result(self):
        return self.node_to_sym, self.sym_to_node

    @functools.lru_cache(maxsize=None)
    def tuple_memo(self, *args):
        return tuple(args)

    def do_e(self, e):
        if (
            isinstance(e, LoopIR.BinOp)
            and (e.op == "%" or e.op == "/")
            and e.type.is_indexable()
        ):
            # found a mod-div site
            tuple_node = self.tupleify(e)
            if tuple_node is None:
                raise UnificationError(
                    f"{e.srcinfo}: cannot handle this '{e.op}' operation"
                )

            # either we have already seen this expression
            if tuple_node in self.tuple_to_sym:
                sym = self.tuple_to_sym[tuple_node]
            # or we are encountering it for the first time
            else:
                opname = "mod" if e.op == "%" else "div"
                node_name = sanitize_str(str(e))
                sym = Sym(f"{opname}_{self.unq_count}_{node_name}")
                self.unq_count += 1
                self.tuple_to_sym[tuple_node] = sym
                self.sym_to_node[sym] = e

            # regardless, record which symbol we've assigned to
            # this specific AST node sub-tree
            self.node_to_sym[id(e)] = sym

        else:
            super().do_e(e)

    def tupleify(self, e):
        if isinstance(e, LoopIR.Read):
            assert len(e.idx) == 0
            if e.name not in self.FV:
                raise UnificationError(
                    f"{e.srcinfo}: "
                    f"Found bound variable '{e.name}' "
                    f"inside of mod or div operation"
                )
            else:
                return e.name
        elif isinstance(e, LoopIR.Const):
            return e.val
        elif isinstance(e, LoopIR.USub):
            return self.tuple_memo("-", self.tupleify(e.arg))
        elif isinstance(e, LoopIR.BinOp):
            return self.tuple_memo(e.op, self.tupleify(e.lhs), self.tupleify(e.rhs))
        else:
            assert False, "Bad case tupleify"

    def do_eff(self, eff):
        return


class BufVar:
    def __init__(self, name, typ, use_win=True):
        self.name = name
        self.typ = typ
        self.n_dim = len(typ.shape())

        self.solution_buf = None

        self.use_win = use_win
        # `win_dim` is the number of dimensions that are
        # point-accesses in a window expression
        # `n_dim` is the number of dimensions that are
        # interval/slice-accesses in a window expression
        # The sum of the two is the size of whatever buffer is unified against
        self.win_dim = None
        self.cases = []
        self.case_var = None

    def set_buf_solution(self, solution_buf):
        self.solution_buf = solution_buf

    def set_window_dim(self, win_dim):
        assert win_dim is not None
        do_setup = self.win_dim is None
        self.win_dim = win_dim

        if do_setup:
            self.cases = []
            self.case_var = Sym(f"{self.name}_which_case")
            full_dim = self.n_dim + win_dim

            for case_id, pt_idxs in enumerate(
                itertools.combinations(range(0, full_dim), win_dim)
            ):
                case_name = (
                    f"{self.name}_"
                    + "_".join([str(i) for i in pt_idxs])
                    + f"_cs{case_id}"
                )

                def make_pt(i):
                    return Sym(f"{case_name}_pt{i}")

                def make_interval(i):
                    return (Sym(f"{case_name}_lo{i}"), Sym(f"{case_name}_hi{i}"))

                idx_vars = [
                    (make_pt(i) if i in pt_idxs else make_interval(i))
                    for i in range(0, full_dim)
                ]

                self.cases.append(idx_vars)

    def get_sz_eq(self, UObj):
        assert self.win_dim is not None
        results = []
        for c in self.cases:
            intervals = [i for i in c if not isinstance(i, Sym)]
            assert len(intervals) == self.n_dim
            for (lo, hi), sz in zip(intervals, self.typ.shape()):
                diff = UEq.Add(UEq.Var(hi), UEq.Scale(-1, UEq.Var(lo)))
                results += [UEq.Eq(diff, UObj.to_ueq(sz))]
        return results

    def all_syms(self):
        if not self.case_var:
            return []
        else:
            xs = [self.case_var]
            for c in self.cases:
                for i in c:
                    if isinstance(i, Sym):
                        xs.append(i)
                    else:
                        xs.append(i[0])
                        xs.append(i[1])
            return xs

    def get_solution(self, UObj, ueq_solutions, srcinfo):
        buf = self.solution_buf
        buf_typ = UObj.FV[buf]
        if not self.case_var:
            return LoopIR.Read(buf, [], buf_typ, srcinfo)
        else:
            which_case = ueq_solutions[self.case_var]
            case = self.cases[which_case]

            def subtract(hi, lo):
                if isinstance(lo, LoopIR.Const) and lo.val == 0:
                    return hi
                else:
                    return LoopIR.BinOp("-", hi, lo, T.index, hi.srcinfo)

            idx = []
            win_shape = []
            for w in case:
                if isinstance(w, Sym):
                    pt = UObj.from_ueq(ueq_solutions[w], srcinfo)
                    idx.append(LoopIR.Point(pt, srcinfo))
                else:
                    lo = UObj.from_ueq(ueq_solutions[w[0]], srcinfo)
                    hi = UObj.from_ueq(ueq_solutions[w[1]], srcinfo)
                    idx.append(LoopIR.Interval(lo, hi, srcinfo))
                    win_shape.append(subtract(hi, lo))

            as_tensor = T.Tensor(win_shape, True, buf_typ.type)
            w_typ = T.Window(buf_typ, as_tensor, buf, idx)
            return LoopIR.WindowExpr(buf, idx, w_typ, srcinfo)


class Unification:
    def __init__(self, subproc, stmt_block, live_vars):
        self.equations = []
        self.stmt_block = stmt_block

        # variables for the UEq system
        # self.holes      = []
        # self.knowns     = []

        # set up different kinds of variables before we
        # begin doing the structural matching...
        # self.arg_syms       = { fa.name : True for fa in subproc.args }
        self.index_holes = [fa.name for fa in subproc.args if fa.type.is_indexable()]
        self.buf_holes = {
            fa.name: BufVar(fa.name, fa.type, fa.type.is_win())
            for fa in subproc.args
            if fa.type.is_numeric()
        }
        self.bool_holes = {fa.name: False for fa in subproc.args if fa.type == T.bool}

        self.stride_holes = {
            fa.name: False for fa in subproc.args if fa.type == T.stride
        }

        # keep track of all buffer names we might need to unify,
        # not just the unknown arguments, but also temporary allocated bufs
        # these variables should ONLY occur on the sub-procedure side
        # of the unification; no BufVars for the original code.
        self.buf_unknowns = self.buf_holes.copy()

        # get the free variables, and lookup their types
        # as well as expanding the free variable set to
        # account for dependent typing
        FV_set = FreeVars(stmt_block).result()
        self.FV = dict()

        def add_fv(x):
            assert x in live_vars, f"expected FV {x} to be live"
            typ = live_vars[x]
            self.FV[x] = typ

            def expand_e(e):
                if isinstance(e, LoopIR.Read):
                    add_fv(e.name)
                elif isinstance(e, LoopIR.USub):
                    expand_e(e.arg)
                elif isinstance(e, LoopIR.BinOp):
                    expand_e(e.lhs)
                    expand_e(e.rhs)

            if isinstance(typ, T.Tensor):
                for e in typ.hi:
                    expand_e(e)
            elif isinstance(typ, T.Window):
                for w in typ.idx:
                    if isinstance(w, LoopIR.Interval):
                        expand_e(w.lo)
                        expand_e(w.hi)
                    else:
                        expand_e(w.pt)
                add_fv(typ.src_buf)

        for x in FV_set:
            add_fv(x)

        # block-side buffer types
        self.bbuf_types = {x: self.FV[x] for x in self.FV if self.FV[x].is_numeric()}

        # self.node_syms  = None
        # self.sym_nodes  = None
        self.node_syms, self.sym_nodes = _Find_Mod_Div_Symbols(
            stmt_block, self.FV
        ).result()

        # substitutions to do of intermediate indexing variables
        self.idx_subst = dict()

        # TODO: Asserts
        # We don't have inequality in EQs IR

        # TODO: Size
        # Inequality...??

        # build up the full set of equations...
        self.unify_stmts(subproc.body, stmt_block)

        # setup the problem
        for nm in self.buf_holes:
            if self.buf_holes[nm].solution_buf is None:
                raise UnificationError(
                    f"Cannot perform unification due to an un-unused argument: {nm}"
                )

        holes = self.index_holes + [
            x for nm in self.buf_holes for x in self.buf_holes[nm].all_syms()
        ]
        knowns = [nm for nm in self.sym_nodes] + [
            nm for nm in self.FV if self.FV[nm].is_indexable()
        ]
        ueq_prob = UEq.problem(holes, knowns, self.equations)

        # solve the problem
        solutions = ueq_prob.solve()
        if solutions is None:
            raise UnificationError(f"Unification of various index expressions failed")

        # construct the solution arguments
        def get_arg(fa):
            if fa.type.is_indexable():
                return self.from_ueq(solutions[fa.name], stmt_block[0].srcinfo)
            elif fa.type == T.bool:
                if fa.name in self.bool_holes:
                    if self.bool_holes[fa.name] is False:
                        return LoopIR.Const(False, T.bool, stmt_block[0].srcinfo)
                    else:
                        return self.bool_holes[fa.name]
                else:
                    # subprocedure argument must be consistent
                    assert False, "bad case"
            elif fa.type == T.stride:
                if fa.name in self.stride_holes:
                    return self.stride_holes[fa.name]
                else:
                    raise UnificationError(f"stride argument {fa.name}" + " unused")
            else:
                assert fa.type.is_numeric()
                bufvar = self.buf_holes[fa.name]
                return bufvar.get_solution(self, solutions, stmt_block[0].srcinfo)

        self.new_args = [get_arg(fa) for fa in subproc.args]

    def err(self):
        raise TypeError("subproc and pattern don't match")

    def result(self):
        return self.new_args

    # ----------

    def to_ueq(self, e, in_subproc=False):
        insp = in_subproc
        if isinstance(e, LoopIR.Read):
            assert len(e.idx) == 0
            name = self.idx_subst[e.name] if e.name in self.idx_subst else e.name
            return UEq.Var(name)
        elif isinstance(e, LoopIR.Const):
            return UEq.Const(e.val)
        elif isinstance(e, LoopIR.USub):
            return UEq.Scale(-1, self.to_ueq(e.arg, insp))
        elif isinstance(e, LoopIR.BinOp):
            if e.op == "+":
                return UEq.Add(self.to_ueq(e.lhs, insp), self.to_ueq(e.rhs, insp))
            elif e.op == "-":
                rhs = UEq.Scale(-1, self.to_ueq(e.rhs, insp))
                return UEq.Add(self.to_ueq(e.lhs, insp), rhs)
            elif e.op == "*":
                if isinstance(e.lhs, LoopIR.Const):
                    return UEq.Scale(e.lhs.val, self.to_ueq(e.rhs, insp))
                elif isinstance(e.rhs, LoopIR.Const):
                    return UEq.Scale(e.rhs.val, self.to_ueq(e.lhs, insp))
                else:
                    assert False, "unexpected multiplication; improve the code here"
            elif e.op == "/" or e.op == "%":
                if in_subproc:
                    raise UnificationError(
                        f"unification with sub-procedures making use of "
                        f"'%' or '/' operations is not currently supported"
                    )
                else:
                    name = self.node_syms[id(e)]
                    return UEq.Var(name)
            else:
                assert False, f"bad op case: {e.op}"
        else:
            assert False, "unexpected affine expression case"

    def from_ueq(self, e, srcinfo=null_srcinfo()):
        if isinstance(e, UEq.Var):
            if e.name in self.sym_nodes:
                return self.sym_nodes[e.name]
            else:
                typ = self.FV[e.name]
                return LoopIR.Read(e.name, [], typ, srcinfo)

        elif isinstance(e, UEq.Const):
            return LoopIR.Const(e.val, T.int, srcinfo)
        elif isinstance(e, UEq.Add):
            lhs = self.from_ueq(e.lhs, srcinfo)
            rhs = self.from_ueq(e.rhs, srcinfo)
            typ = (
                lhs.type
                if rhs.type == T.int
                else rhs.type
                if lhs.type == T.int
                else lhs.type
                if rhs.type == T.size
                else rhs.type
            )
            return LoopIR.BinOp("+", lhs, rhs, typ, srcinfo)
        elif isinstance(e, UEq.Scale):
            lhs = LoopIR.Const(e.coeff, T.int, srcinfo)
            rhs = self.from_ueq(e.e, srcinfo)
            return LoopIR.BinOp("*", lhs, rhs, rhs.type, srcinfo)
        else:
            assert False, "bad case"

    # ----------

    def all_bound_e(self, be):
        if isinstance(be, LoopIR.Read):
            if be.name not in self.FV:
                return False
            return all(self.all_bound_e(i) for i in be.idx)
        elif isinstance(be, LoopIR.Const):
            return True
        elif isinstance(be, LoopIR.USub):
            return self.all_bound_e(be.arg)
        elif isinstance(be, LoopIR.BinOp):
            return self.all_bound_e(be.lhs) and self.all_bound_e(be.rhs)
        elif isinstance(be, LoopIR.BuiltIn):
            return all(self.all_bound_e(a) for a in be.args)
        else:
            assert False, "unsupported case"

    def is_exact_e(self, e0, e1):
        if type(e0) is not type(e1):
            return False
        elif isinstance(e0, LoopIR.Read):
            return e0.name == e1.name and all(
                self.is_exact_e(i0, i1) for i0, i1 in zip(e0.idx, e1.idx)
            )
        elif isinstance(e0, LoopIR.Const):
            return e0.val == e1.val
        elif isinstance(e0, LoopIR.USub):
            return self.is_exact_e(e0.arg, e1.arg)
        elif isinstance(e0, LoopIR.BinOp):
            return (
                e0.op == e1.op
                and self.is_exact_e(e0.lhs, e1.lhs)
                and self.is_exact_e(e0.rhs, e1.rhs)
            )
        elif isinstance(e0, LoopIR.BuiltIn):
            return e0.f == e1.f and all(
                self.is_exact_e(a0, a1) for a0, a1 in zip(e0.args, e1.args)
            )
        else:
            assert False, "unsupported case"

    # ----------

    def unify_affine_e(self, pa, ba):
        self.equations.append(UEq.Eq(self.to_ueq(pa, in_subproc=True), self.to_ueq(ba)))

    def unify_bool_hole(self, pe, be):
        assert pe.type == be.type == T.bool
        assert isinstance(pe, LoopIR.Read) and pe.name in self.bool_holes

        if not self.all_bound_e(be):
            raise UnificationError(
                f"Cannot unify expression {be} (@{be.srcinfo}) with the "
                f"boolean argument {pe.name} because it contains "
                f"variables that are not free in the code being replaced"
            )

        # if we haven't yet unified this name with an expression
        lookup = self.bool_holes[pe.name]
        if lookup is False:
            self.bool_holes[pe.name] = be
        elif not self.is_exact_e(be, lookup):
            raise UnificationError(
                f"Cannot unify the boolean argument {pe.name} with two "
                f"seemingly inequivalent expressions "
                f"{be} (@{be.srcinfo}) and {lookup} (@{lookup.srcinfo})"
            )

    def unify_stride_hole(self, pe, be):
        assert pe.type == be.type == T.stride
        assert isinstance(pe, LoopIR.Read) and pe.name in self.stride_holes

        # TODO: Add checks here??
        lookup = self.stride_holes[pe.name]
        if lookup is False:
            self.stride_holes[pe.name] = be

    def unify_stmts(self, proc_s, block_s):
        if len(proc_s) != len(block_s):
            ploc, bloc = "", ""
            if len(proc_s) > 0:
                ploc = f" (@{proc_s[0].srcinfo})"
            if len(block_s) > 0:
                bloc = f" (@{block_s[0].srcinfo})"
            raise UnificationError(
                f"cannot unify {len(proc_s)} statement(s){ploc} with "
                f"{len(block_s)} statement(s){bloc}"
            )
        elif len(proc_s) == 0:
            return

        ps, proc_s = proc_s[0], proc_s[1:]
        bs, block_s = block_s[0], block_s[1:]

        if type(ps) is not type(bs):
            raise UnificationError(
                f"cannot unify a {type(ps)} statement (@{ps.srcinfo}) with "
                f"a {type(bs)} statement (@{bs.srcinfo})"
            )
        elif isinstance(ps, (LoopIR.Assign, LoopIR.Reduce)):
            self.unify_e(ps.rhs, bs.rhs)
            self.unify_accesses(ps, bs)
        elif isinstance(ps, LoopIR.WriteConfig):
            if ps.config != bs.config or ps.field != bs.field:
                raise UnificationError(
                    f"cannot unify Writeconfig '{pe.config.name()}.{pe.field}' "
                    f"with Writeconfig '{be.config.name()}.{be.field}'"
                )
            self.unify_e(ps.rhs, bs.rhs)
        elif isinstance(ps, LoopIR.Pass):
            pass
        elif isinstance(ps, LoopIR.If):
            self.unify_e(ps.cond, bs.cond)
            self.unify_stmts(ps.body, bs.body)
            self.unify_stmts(ps.orelse, bs.orelse)
        elif isinstance(ps, LoopIR.For):
            # BINDING
            self.idx_subst[ps.iter] = bs.iter
            self.unify_e(ps.lo, bs.lo)
            self.unify_e(ps.hi, bs.hi)
            self.unify_stmts(ps.body, bs.body)
        elif isinstance(ps, LoopIR.Alloc):
            # introduce BufVars on the sub-procedure side of unification
            # and immediately force the solution to match the name found
            # on the original code side of unification
            pvar = BufVar(ps.name, ps.type, use_win=False)
            pvar.set_buf_solution(bs.name)
            self.buf_unknowns[ps.name] = pvar
            self.bbuf_types[bs.name] = bs.type
            self.unify_types(ps.type, bs.type, ps, bs)
        elif isinstance(ps, LoopIR.Call):
            if ps.f != bs.f:
                raise UnificationError(
                    f"cannot unify a call to '{ps.f.name()}' (@{ps.srcinfo}) "
                    f"with a call to {bs.f.name()} (@{bs.srcinfo})"
                )
            for pe, be in zip(ps.args, bs.args):
                self.unify_e(pe, be)
        elif isinstance(ps, LoopIR.WindowStmt):
            self.unify_e(ps.rhs, bs.rhs)
            # new name identification is similar to Alloc
            pvar = BufVar(ps.name, ps.rhs.type.as_tensor, use_win=False)
            pvar.set_buf_solution(bs.name)
            self.buf_unknowns[ps.name] = pvar
            self.bbuf_types[bs.name] = bs.rhs.type.as_tensor

        # tail recursion
        self.unify_stmts(proc_s, block_s)

    # directly unify two buffer names without adding any windowing
    def unify_buf_name_no_win(self, pname, bname):
        pvar = self.buf_unknowns[pname]

        if pvar.use_win:
            raise UnificationError(
                f"Cannot unify the windowed buffer '{pname}' "
                f"with the buffer '{bname}' because '{bname}' is used "
                f"in a position where windowing is not currently supported"
            )

        if pvar.solution_buf and pvar.solution_buf != bname:
            raise UnificationError(
                f"The buffer {pname} cannot be unified to both "
                f"the buffer {pvar.solution_buf}, and the buffer {bname}"
            )
        else:
            pvar.set_buf_solution(bname)

    def unify_accesses(self, pnode, bnode):
        pbuf, pidx = pnode.name, pnode.idx
        bbuf, bidx = bnode.name, bnode.idx
        pvar = self.buf_unknowns[pbuf]

        idx_gap = len(bidx) - len(pidx)
        # first, reject any numbers of indices that absolutely
        # cannot be made to work
        if idx_gap < 0:
            raise UnificationError(
                f"the access to '{pbuf}' (@{pnode.srcinfo}) has too many "
                f"indices ({len(pidx)}, compared to {len(bidx)}) to unify with "
                f"the access to '{bbuf}' (@{bnode.srcinfo})"
            )

        # handle special case of unindexed buffers used in
        # call-argument position
        if isinstance(bnode, LoopIR.Read) and len(bnode.type.shape()) > 0:
            assert len(bidx) == 0
            # we now know that bnode looks something like `x` where
            # `x` is not a scalar
            if len(pnode.type.shape()) == 0:
                raise UnificationError(
                    f"Could not unify buffer '{pbuf}' (@{pnode.srcinfo}) "
                    f"with buffer '{bbuf}' (@{bnode.srcinfo})"
                )
            else:
                assert len(pidx) == 0
                self.unify_types(pnode.type, bnode.type, pnode, bnode)
                self.unify_buf_name_no_win(pbuf, bbuf)
                return
        elif isinstance(pnode, LoopIR.Read) and len(pnode.type.shape()) > 0:
            # NOTE: bnode is not trivial b/c of the elif
            raise UnificationError(
                f"Unification of the simple call argument "
                f"'{pbuf}' (@{pnode.srcinfo}) "
                f"with a non-simple call argument "
                f"'{bbuf}' (@{bnode.srcinfo}) "
                f"is currently unsupported"
            )

        # otherwise, we can be sure that everything has been
        # accessed all the way down to a particular scalar value
        assert pnode.type.is_real_scalar() and bnode.type.is_real_scalar()

        # How to unify accesses when there is no intermediate windowing
        if not pvar.use_win:
            if idx_gap == 0:
                # with the index gap closed...
                for pi, bi in zip(pidx, bidx):
                    self.unify_affine_e(pi, bi)
                self.unify_types(pvar.typ, self.bbuf_types[bbuf], pnode, bnode)
                self.unify_buf_name_no_win(pbuf, bbuf)
            elif len(pidx) == 0 and self.is_proc_constant(bbuf):
                raise NotImplementedError("Unify buffer access with constant")
            else:
                raise UnificationError(
                    f"cannot unify the access to '{pbuf}' (@{pnode.srcinfo}) "
                    f"using {len(pidx)} indices with the access to "
                    f"'{bbuf}' (@{bnode.srcinfo}) using {len(bidx)} indices."
                )

        # Otherwise, how to unify accesses WITH windowing in the way
        else:
            if pvar.win_dim is not None and pvar.win_dim != idx_gap:
                raise UnificationError(
                    f"cannot unify '{pbuf}' (@{pnode.srcinfo}) "
                    f"with '{bbuf}' (@{bnode.srcinfo}) "
                    f"because '{pbuf}' is already being windowed down "
                    f"from a {pvar.n_dim + pvar.win_dim} dimension tensor, "
                    f"but is required to be windowed down from a "
                    f"{len(bidx)} dimension tensor here"
                )

            # set up all the case variables and the
            # equations relating windowing lo/hi expressions to
            # the resulting window's size-type expressions
            # Guard this to prevent redundant imposition of sizing equations
            if pvar.win_dim is None:
                pvar.set_buf_solution(bbuf)
                pvar.set_window_dim(idx_gap)
                self.equations += pvar.get_sz_eq(self)

            # now construct the equations relating the indexing on
            # the two sides of this access in all possible cases
            def case_conj(case_idxs):
                eqs = []
                tmp_pidx = pidx.copy()
                assert len(bidx) == len(case_idxs)
                for bi, wi in zip(bidx, case_idxs):
                    be = self.to_ueq(bi)
                    if isinstance(wi, Sym):  # point access from window
                        pe = UEq.Var(wi)
                    else:  # interval access
                        pe = UEq.Add(UEq.Var(wi[0]), self.to_ueq(tmp_pidx.pop(0)))
                    eqs.append(UEq.Eq(pe, be))

                assert len(tmp_pidx) == 0
                return UEq.Conj(eqs)

            cases = UEq.Cases(pvar.case_var, [case_conj(cidxs) for cidxs in pvar.cases])
            self.equations.append(cases)

    def unify_types(self, pt, bt, pnode, bnode):
        if pt.is_real_scalar() and bt.is_real_scalar():
            return  # success
        elif pt.is_indexable() and bt.is_indexable():
            return  # success
        elif pt == T.bool and bt == T.bool:
            return  # success
        elif pt.is_tensor_or_window() and bt.is_tensor_or_window():
            if len(pt.shape()) != len(bt.shape()):
                raise UnificationError(
                    f"cannot unify a tensor-type of "
                    f"{len(pt.shape())} dimensions (@{pnode.srcinfo}) with "
                    f"a tensor-type of {len(bt.shape())} dimensions "
                    f"(@{bnode.srcinfo})"
                )
            for psz, bsz in zip(pt.shape(), bt.shape()):
                self.unify_affine_e(psz, bsz)
        else:
            raise UnificationError(
                f"cannot unify type {pt} (@{pnode.srcinfo}) with "
                f"type {bt} (@{bnode.srcinfo})"
            )

    @staticmethod
    def comparision_to_unification_expr(e):
        def sub(lhs, rhs):
            return LoopIR.BinOp("-", lhs, rhs, T.index, null_srcinfo())

        def add1(e):
            one = LoopIR.Const(1, T.int, null_srcinfo())
            return LoopIR.BinOp("+", e, one, T.index, null_srcinfo())

        if e.op == "<" or e.op == "==":
            # rewrite `lhs op rhs` into `0 op rhs - lhs`
            return sub(e.rhs, e.lhs)
        elif e.op == "<=":
            # rewrite `lhs <= rhs` into `0 < rhs - lhs + 1`
            return add1(sub(e.rhs, e.lhs))
        elif e.op == ">":
            # rewrite `lhs > rhs` into `0 < lhs - rhs`
            return sub(e.lhs, e.rhs)
        elif e.op == ">=":
            # rewrite `lhs >= rhs` into `0 < lhs - rhs + 1`
            return add1(sub(e.lhs, e.rhs))
        else:
            assert False, f"Unreachable op found: {e.op}"

    def unify_e(self, pe, be):
        if pe.type.is_indexable() != be.type.is_indexable() or (pe.type == T.bool) != (
            be.type == T.bool
        ):
            raise UnificationError(
                f"expected expressions to have similar types:\n"
                f"  {pe}: {pe.type} [{pe.srcinfo}]\n"
                f"  {be}: {be.type} [{be.srcinfo}]"
            )
        elif pe.type.is_indexable():
            # convert to an equality
            self.unify_affine_e(pe, be)
            return
        elif (
            pe.type == T.bool
            and isinstance(pe, LoopIR.Read)
            and pe.name in self.bool_holes
        ):
            self.unify_bool_hole(pe, be)
            return
        elif (
            pe.type == T.stride
            and isinstance(pe, LoopIR.Read)
            and pe.name in self.stride_holes
        ):
            self.unify_stride_hole(pe, be)
            return

        if type(pe) is not type(be):
            raise UnificationError(
                f"cannot unify a {type(pe)} expression (@{pe.srcinfo}) with "
                f"a {type(be)} expression (@{be.srcinfo})"
            )
        elif isinstance(pe, LoopIR.Read):
            assert pe.type.is_numeric(), "unhandled expression type...?"
            self.unify_accesses(pe, be)
        elif isinstance(pe, LoopIR.Const):
            if pe.val != be.val:
                raise UnificationError(
                    f"cannot unify {pe.val} (@{pe.srcinfo}) with "
                    f"{be.val} (@{be.srcinfo})"
                )
        elif isinstance(pe, LoopIR.USub):
            self.unify_e(pe.arg, be.arg)
        elif isinstance(pe, LoopIR.BinOp):
            exprs = [pe.rhs, pe.lhs, be.rhs, be.lhs]
            if pe.op in comparision_ops and all(e.type.is_indexable() for e in exprs):
                inequality_ops = comparision_ops - {"=="}
                if pe.op == be.op or (
                    pe.op in inequality_ops and be.op in inequality_ops
                ):
                    pe_e = self.comparision_to_unification_expr(pe)
                    be_e = self.comparision_to_unification_expr(be)
                    self.unify_e(pe_e, be_e)
                    return
            if pe.op != be.op:
                raise UnificationError(
                    f"cannot unify a '{pe.op}' (@{pe.srcinfo}) with "
                    f"a '{be.op}'' (@{be.srcinfo})"
                )
            self.unify_e(pe.lhs, be.lhs)
            self.unify_e(pe.rhs, be.rhs)
        elif isinstance(pe, LoopIR.BuiltIn):
            if pe.f != be.f:
                raise UnificationError(
                    f"cannot unify builtin '{pe.f.name()}' (@{pe.srcinfo}) "
                    f"with builtin '{be.f.name()}'' (@{be.srcinfo})"
                )
            for pa, ba in zip(pe.args, be.args):
                self.unify_e(pa, ba)
        elif isinstance(pe, LoopIR.ReadConfig):
            if pe.config != be.config or pe.field != be.field:
                raise UnificationError(
                    f"cannot unify readconfig '{pe.config.name()}.{pe.field}' "
                    f"with readconfig '{be.config.name()}.{be.field}'"
                )
        elif isinstance(pe, LoopIR.WindowExpr):
            pvar = self.buf_unknowns[pe.name]

            # unify the two buffers
            self.unify_buf_name_no_win(pe.name, be.name)
            self.unify_types(pvar.typ, self.bbuf_types[be.name], pe, be)

            # unify the two windowing expressions
            if len(pe.idx) != len(be.idx):
                raise UnificationError(
                    f"cannot unify the windowing of {pe.name} (@{pe.srcinfo}) "
                    f"using {len(pe.idx)} indices with the windowing of "
                    f"{be.name} (@{be.srcinfo}) using {len(be.idx)}"
                )

            for i, (pw, bw) in enumerate(zip(pe.idx, be.idx)):
                if type(pw) is not type(bw):
                    raise UnificationError(
                        f"cannot unify the windowing of "
                        f"{pe.name} (@{pe.srcinfo}) with the windowing of "
                        f"{be.name} (@{be.srcinfo}) because one evaluates to a "
                        f"point at index {i}, while the other evaluates to an "
                        f"interval"
                    )
                elif isinstance(pw, LoopIR.Point):
                    self.unify_affine_e(pw.pt, bw.pt)
                else:
                    self.unify_affine_e(pw.lo, bw.lo)
                    self.unify_affine_e(pw.hi, bw.hi)
        else:
            assert False, f"bad case of {type(pe)}"

    def is_proc_constant(self, bbuf):
        deps = LoopIR_Dependencies(bbuf, self.stmt_block).result()
        return all(dep in self.FV for dep in deps)



from __future__ import annotations

import numpy as np
import pytest

try:
    import torch
    import torch.nn.functional as F
except ImportError:
    pytest.skip("pytorch is not available, skipping winograd", allow_module_level=True)

from exo import proc


def wconv_3x3():
    # res <- A @ B @ A^T
    @proc
    def multiply(n: size, m: size, A: f32[n, m], B: f32[m, m], res: f32[n, n]):

        tmp: f32[n, m]
        for i in seq(0, n):
            for j in seq(0, m):
                tmp[i, j] = 0.0
                for k in seq(0, m):
                    tmp[i, j] += A[i, k] * B[k, j]

        for i in seq(0, n):
            for j in seq(0, n):
                res[i, j] = 0.0
                for k in seq(0, m):
                    res[i, j] += tmp[i, k] * A[j, k]

                    # C <- A @ B (element wise)

    @proc
    def element(n: size, A: f32[n, n], B: f32[n, n], C: f32[n, n]):

        for i in seq(0, n):
            for j in seq(0, n):
                C[i, j] = A[i, j] * B[i, j]

    @proc
    def wconv(
        inp: f32[4, 4],
        kernel: f32[3, 3],
        res: f32[2, 2],
        B_T: f32[4, 4],
        G: f32[4, 3],
        A_T: f32[2, 4],
    ):

        # Cannot do this
        # B_T = [1,	0,	-1,	0, \
        #       0,	1,	1,	0, \
        #       0,	-1,	1,	0, \
        #       0,	1,	0,	-1]

        U: f32[4, 4]
        # U <- G @ kernel @ G^T
        multiply(4, 3, G, kernel, U)

        V: f32[4, 4]
        # V <- B^T @ inp @ B
        multiply(4, 4, B_T, inp, V)

        M: f32[4, 4]
        # M <- U @ V (element wise)
        element(4, U, V, M)

        # res <- A^T @ M @ A
        multiply(2, 4, A_T, M, res)

    return wconv


def test_winograd(compiler):
    conv = wconv_3x3()

    wconv = compiler.compile(conv)

    B_T = np.array([(1, 0, -1, 0), (0, 1, 1, 0), (0, -1, 1, 0), (0, 1, 0, -1)])

    G = np.array([(1, 0, 0), (1 / 2, 1 / 2, 1 / 2), (1 / 2, -1 / 2, 1 / 2), (0, 0, 1)])

    A_T = np.array([(1, 1, 1, 0), (0, 1, -1, -1)])

    n = 4
    m = 3
    # batch, channel = 1
    kernel = torch.randint(high=100, size=(1, 1, m, m)).float()
    inp = torch.randint(high=100, size=(1, 1, n, n)).float()

    ref = F.conv2d(inp, kernel)

    kernel = kernel.detach().numpy()
    inp = inp.detach().numpy()
    res = np.zeros(shape=(2, 2), dtype=np.float32)

    wconv(None, inp, kernel, res, B_T, G, A_T)
    print(res)

    # TODO: finish test, validate output (returns NaN right now)


"""
    n_size = IMAGE.shape[0]
    m_size = IMAGE.shape[1]
    k_size = 5
    kernel = gkern(k_size, 1)
    res = nprand(size=(n_size, m_size))
    res_c = cvt_c(res)
    test_lib = generate_lib(filename)
    test_lib.blur(POINTER(c_int)(), c_int(n_size), c_int(m_size), c_int(
        k_size), cvt_c(IMAGE), cvt_c(kernel), res_c)
    res_c = np.ctypeslib.as_array(res_c, shape=(n_size, m_size))
    res_c = res_c.astype(np.uint8)
    out = Image.fromarray(res_c)
    out.save(os.path.join(TMP_DIR, filename + '_out.png'))
"""
